"""Class for rendering with Skia"""

import dataclasses
import typing
import typing_extensions
import math
import os

try:
    import skia
except ModuleNotFoundError as e:
    raise Exception(
        "You might want to install momapy with the skia extra: momapy[skia]"
    ) from e

import momapy.drawing
import momapy.geometry
import momapy.rendering.core


@dataclasses.dataclass(kw_only=True)
class SkiaRenderer(momapy.rendering.core.StatefulRenderer):
    """Class for skia renderers"""

    formats: typing.ClassVar[list[str]] = ["pdf", "svg", "png", "jpeg", "webp"]
    _de_class_func_mapping: typing.ClassVar[dict] = {
        momapy.drawing.Group: "_render_group",
        momapy.drawing.Path: "_render_path",
        momapy.drawing.Text: "_render_text",
        momapy.drawing.Ellipse: "_render_ellipse",
        momapy.drawing.Rectangle: "_render_rectangle",
    }
    _pa_class_func_mapping: typing.ClassVar[dict] = {
        momapy.drawing.MoveTo: "_add_move_to",
        momapy.drawing.LineTo: "_add_line_to",
        momapy.drawing.CurveTo: "_add_curve_to",
        momapy.drawing.QuadraticCurveTo: "_add_quadratic_curve_to",
        momapy.drawing.ClosePath: "_add_close_path",
        momapy.drawing.EllipticalArc: "_add_elliptical_arc",
    }
    _tr_class_func_mapping: typing.ClassVar[dict] = {
        momapy.geometry.Translation: "_add_translation",
        momapy.geometry.Rotation: "_add_rotation",
        momapy.geometry.Scaling: "_add_scaling",
        momapy.geometry.MatrixTransformation: "_add_matrix_transformation",
    }
    _fe_class_func_mapping: typing.ClassVar[dict] = {
        momapy.drawing.DropShadowEffect: "_make_drop_shadow_effect",
        momapy.drawing.CompositeEffect: "_make_composite_effect",
        momapy.drawing.GaussianBlurEffect: "_make_gaussian_blur_effect",
        momapy.drawing.OffsetEffect: "_make_offset_effect",
        momapy.drawing.FloodEffect: "_make_flood_effect",
    }
    _fe_composite_comp_op_blendmode_mapping: typing.ClassVar[dict] = {
        momapy.drawing.CompositionOperator.OVER: skia.BlendMode.kSrcOver,
        momapy.drawing.CompositionOperator.IN: skia.BlendMode.kSrcIn,
        momapy.drawing.CompositionOperator.OUT: skia.BlendMode.kSrcOut,
        momapy.drawing.CompositionOperator.ATOP: skia.BlendMode.kSrcATop,
        momapy.drawing.CompositionOperator.XOR: skia.BlendMode.kXor,
        momapy.drawing.CompositionOperator.LIGHTER: skia.BlendMode.kLighten,
        momapy.drawing.CompositionOperator.ARTIHMETIC: None,
    }
    _fe_gaussian_blur_edgemode_tilemode_mapping: typing.ClassVar[dict] = {
        momapy.drawing.EdgeMode.WRAP: skia.TileMode.kMirror,
        momapy.drawing.EdgeMode.DUPLICATE: skia.TileMode.kClamp,
        None: skia.TileMode.kDecal,
    }
    _te_font_style_slant_mapping: typing.ClassVar[dict] = {
        momapy.drawing.FontStyle.NORMAL: skia.FontStyle.Slant.kUpright_Slant,
        momapy.drawing.FontStyle.ITALIC: skia.FontStyle.Slant.kItalic_Slant,
        momapy.drawing.FontStyle.OBLIQUE: skia.FontStyle.Slant.kOblique_Slant,
    }
    canvas: skia.Canvas = dataclasses.field(metadata={"description": "A skia canvas"})
    _config: dict = dataclasses.field(default_factory=dict)
    _skia_typefaces: dict = dataclasses.field(default_factory=dict)
    _skia_fonts: dict = dataclasses.field(default_factory=dict)

    @classmethod
    def from_file(
        cls,
        file_path: str | os.PathLike,
        width: float,
        height: float,
        format_: typing.Literal["pdf", "svg", "png", "jpeg", "webp"] = "pdf",
    ) -> typing_extensions.Self:
        config = {}
        if format_ == "pdf":
            stream = skia.FILEWStream(file_path)
            document = skia.PDF.MakeDocument(stream)
            canvas = document.beginPage(width, height)
            config["stream"] = stream
            config["document"] = document
        elif format_ in ["png", "jpeg", "webp"]:
            surface = skia.Surface(width=int(width), height=int(height))
            canvas = surface.getCanvas()
            config["surface"] = surface
            config["file_path"] = file_path
        elif format_ == "svg":
            stream = skia.FILEWStream(file_path)
            canvas = skia.SVGCanvas.Make((width, height), stream)
            config["stream"] = stream
        config["file_path"] = file_path
        config["width"] = width
        config["height"] = height
        config["format"] = format_
        return cls(canvas=canvas, _config=config)

    def begin_session(self):
        pass

    def end_session(self):
        self.canvas.flush()
        format_ = self._config.get("format")
        if format_ == "pdf":
            self._config["document"].endPage()
            self._config["document"].close()
        elif format_ == "png":
            image = self._config["surface"].makeImageSnapshot()
            image.save(self._config["file_path"], skia.kPNG)
        elif format_ == "jpeg":
            image = self._config["surface"].makeImageSnapshot()
            image.save(self._config["file_path"], skia.kJPEG)
        elif format_ == "webp":
            image = self._config["surface"].makeImageSnapshot()
            image.save(self._config["file_path"], skia.kWEBP)
        elif format_ == "svg":
            del self.canvas
            self._config["stream"].flush()

    def new_page(self, width, height):
        format_ = self._config.get("format")
        if format_ == "pdf":
            self._config["document"].endPage()
            canvas = self._config["document"].beginPage(width, height)
            self.canvas = canvas

    def render_map(self, map_):
        self.render_layout_element(map_.layout)

    def render_layout_element(self, layout_element):
        drawing_elements = layout_element.drawing_elements()
        for drawing_element in drawing_elements:
            self.render_drawing_element(drawing_element)

    def render_drawing_element(self, drawing_element):
        self.save()
        self.set_current_state_from_drawing_element(drawing_element)
        self._add_transform_from_drawing_element(drawing_element)
        class_ = type(drawing_element)
        if issubclass(class_, momapy.builder.Builder):
            class_ = class_._cls_to_build
        de_func = getattr(self, self._de_class_func_mapping[class_])
        filter = self.get_current_value("filter")
        if filter is not momapy.drawing.NoneValue:
            bbox = drawing_element.bbox()
            saved_canvas = self.canvas
            recorder = skia.PictureRecorder()
            canvas = recorder.beginRecording(
                skia.Rect.MakeXYWH(
                    bbox.north_west().x,
                    bbox.north_west().y,
                    bbox.width,
                    bbox.height,
                )
            )
            self.canvas = canvas
            de_func(drawing_element)
            picture = recorder.finishRecordingAsPicture()
            skia_paint = self._make_filter_paint(
                filter, drawing_element.get_filter_region()
            )
            self.canvas = saved_canvas
            self.canvas.drawPicture(picture, paint=skia_paint)
        else:
            de_func(drawing_element)
        self.restore()

    def self_save(self):
        self.canvas.save()

    def self_restore(self):
        self.canvas.restore()

    def _make_stroke_paint(self):
        if self.get_current_value("stroke_dasharray") is not momapy.drawing.NoneValue:
            skia_path_effect = skia.DashPathEffect.Make(
                list(self.get_current_value("stroke_dasharray")),
                self.get_current_value("stroke_dashoffset"),
            )
        else:
            skia_path_effect = None
        skia_paint = skia.Paint(
            AntiAlias=True,
            Color4f=skia.Color4f(
                self.get_current_value("stroke").to_rgba(rgba_range=(0.0, 1.0))
            ),
            StrokeWidth=self.get_current_value("stroke_width"),
            PathEffect=skia_path_effect,
            Style=skia.Paint.kStroke_Style,
        )
        return skia_paint

    def _make_fill_paint(self):
        skia_paint = skia.Paint(
            AntiAlias=True,
            Color4f=skia.Color4f(
                self.get_current_value("fill").to_rgba(rgba_range=(0.0, 1.0))
            ),
            Style=skia.Paint.kFill_Style,
        )
        return skia_paint

    def _make_filter_paint(self, filter_, filter_region):
        dskia_filters = {}
        for filter_effect in filter_.effects:
            class_ = type(filter_effect)
            if issubclass(class_, momapy.builder.Builder):
                class_ = class_._cls_to_build
            fe_func = getattr(self, self._fe_class_func_mapping[class_])
            skia_filter = fe_func(filter_effect, filter_region, dskia_filters)
            if filter_effect.result is not None:
                dskia_filters[filter_effect.result] = skia_filter
        skia_paint = skia.Paint(AntiAlias=True, ImageFilter=skia_filter)
        return skia_paint

    def _make_crop_rect_from_filter_region(self, filter_region):
        crop_rect = skia.IRect.MakeXYWH(
            round(filter_region.north_west().x),
            round(filter_region.north_west().y),
            round(filter_region.width),
            round(filter_region.height),
        )
        return crop_rect

    def _make_input_filter_from_reference(self, dskia_filters, filter_reference):
        if isinstance(filter_reference, momapy.drawing.FilterEffectInput):
            return None  # all SVG options default to source bitmap in skia
        in_skia_filter = dskia_filters.get(filter_reference)
        if in_skia_filter is None:  # if no reference or bad reference
            if dskia_filters:  # we take the last filter effect primitive if it exists, otherwise remains None (source Bitmap)
                in_skia_filter = dskia_filters[list(dskia_filters.keys())[-1]]
        return in_skia_filter

    def _make_drop_shadow_effect(self, filter_effect, filter_region, dskia_filters):
        crop_rect = self._make_crop_rect_from_filter_region(filter_region)
        skia_filter = skia.ImageFilters.DropShadow(
            dx=filter_effect.dx,
            dy=filter_effect.dy,
            sigmaX=filter_effect.std_deviation,
            sigmaY=filter_effect.std_deviation,
            color=skia.Color4f(
                *filter_effect.flood_color.to_rgb(rgb_range=(0, 1)),
                filter_effect.flood_opacity,
            ),
            cropRect=crop_rect,
        )
        return skia_filter

    def _make_composite_effect(self, filter_effect, filter_region, dskia_filters):
        crop_rect = self._make_crop_rect_from_filter_region(filter_region)
        in_skia_filter = self._make_input_filter_from_reference(
            dskia_filters, filter_effect.in_
        )
        in2_skia_filter = self._make_input_filter_from_reference(
            dskia_filters, filter_effect.in2
        )
        blend_mode = self._fe_composite_comp_op_blendmode_mapping[
            filter_effect.operator
        ]  # TODO: arithmetic operator
        skia_filter = skia.ImageFilters.Xfermode(
            mode=blend_mode,
            background=in2_skia_filter,
            foreground=in_skia_filter,
            cropRect=crop_rect,
        )
        return skia_filter

    def _make_flood_effect(self, filter_effect, filter_region, dskia_filters):
        crop_rect = self._make_crop_rect_from_filter_region(filter_region)
        skia_paint = skia.Paint(
            AntiAlias=True,
            Color4f=skia.Color4f(
                *filter_effect.flood_color.to_rgb(rgb_range=(0, 1)),
                filter_effect.flood_opacity,
            ),
            Style=skia.Paint.kFill_Style,
        )
        skia_filter = skia.ImageFilters.Paint(
            paint=skia_paint,
            cropRect=crop_rect,
        )
        return skia_filter

    def _make_gaussian_blur_effect(self, filter_effect, filter_region, dskia_filters):
        crop_rect = self._make_crop_rect_from_filter_region(filter_region)
        in_skia_filter = self._make_input_filter_from_reference(
            dskia_filters, filter_effect.in_
        )
        tile_mode = self._fe_gaussian_blur_edgemode_tilemode_mapping[
            filter_effect.edge_mode
        ]
        skia_filter = skia.ImageFilters.Blur(
            sigmaX=filter_effect.std_deviation,
            sigmaY=filter_effect.std_deviation,
            tileMode=tile_mode,
            input=in_skia_filter,
            cropRect=crop_rect,
        )
        return skia_filter

    def _make_offset_effect(self, filter_effect, filter_region, dskia_filters):
        crop_rect = self._make_crop_rect_from_filter_region(filter_region)
        in_skia_filter = self._make_input_filter_from_reference(
            dskia_filters, filter_effect.in_
        )
        skia_filter = skia.ImageFilters.Offset(
            dx=filter_effect.dx,
            dy=filter_effect.dy,
            input=in_skia_filter,
            cropRect=crop_rect,
        )
        return skia_filter

    def _add_transform_from_drawing_element(self, drawing_element):
        if (
            drawing_element.transform is not None
            and drawing_element.transform != momapy.drawing.NoneValue
        ):
            for transformation in drawing_element.transform:
                self._add_transformation(transformation)

    def _add_transformation(self, transformation):
        class_ = type(transformation)
        if issubclass(class_, momapy.builder.Builder):
            class_ = class_._cls_to_build
        tr_func = getattr(self, self._tr_class_func_mapping[class_])
        return tr_func(transformation)

    def _render_group(self, group):
        for drawing_element in group.elements:
            self.render_drawing_element(drawing_element)

    def _add_path_action_to_skia_path(self, skia_path, path_action):
        class_ = type(path_action)
        if issubclass(class_, momapy.builder.Builder):
            class_ = class_._cls_to_build
        pa_func = getattr(self, self._pa_class_func_mapping[class_])
        pa_func(skia_path, path_action)

    def _make_skia_path(self, path):
        skia_path = skia.Path()
        for action in path.actions:
            self._add_path_action_to_skia_path(skia_path, action)
        return skia_path

    def _render_path(self, path):
        skia_path = self._make_skia_path(path)
        if self.get_current_value("fill") != momapy.drawing.NoneValue:
            skia_paint = self._make_fill_paint()
            self.canvas.drawPath(path=skia_path, paint=skia_paint)
        if self.get_current_value("stroke") != momapy.drawing.NoneValue:
            skia_paint = self._make_stroke_paint()
            self.canvas.drawPath(path=skia_path, paint=skia_paint)

    def _render_text(self, text):
        font_family = self.get_current_value("font_family")
        font_weight = self.get_current_value("font_weight")
        font_style = self.get_current_value("font_style")
        skia_typeface = self._skia_typefaces.get((font_family, font_weight, font_style))
        if skia_typeface is None:
            skia_font_slant = self._te_font_style_slant_mapping[font_style]
            skia_font_style = skia.FontStyle(
                weight=int(font_weight),
                slant=skia_font_slant,
                width=skia.FontStyle.kNormal_Width,
            )
            skia_typeface = skia.Typeface(
                familyName=font_family,
                fontStyle=skia_font_style,
            )
            self._skia_typefaces[(font_family, font_weight, font_style)] = skia_typeface
        font_size = self.get_current_value("font_size")
        skia_font = self._skia_fonts.get(
            (
                font_family,
                font_weight,
                font_style,
                font_size,
            )
        )
        if skia_font is None:
            skia_font = skia.Font(
                typeface=skia_typeface,
                size=font_size,
            )
            self._skia_fonts[
                (
                    font_family,
                    font_weight,
                    font_style,
                    font_size,
                )
            ] = skia_font
        if self.get_current_value("fill") != momapy.drawing.NoneValue:
            skia_paint = self._make_fill_paint()
            self.canvas.drawString(
                text=text.text,
                x=text.x,
                y=text.y,
                font=skia_font,
                paint=skia_paint,
            )
        if self.get_current_value("stroke") != momapy.drawing.NoneValue:
            skia_paint = self._make_stroke_paint()
            self.canvas.drawString(
                text=text.text,
                x=text.x,
                y=text.y,
                font=skia_font,
                paint=skia_paint,
            )

    def _render_ellipse(self, ellipse):
        skia_rect = skia.Rect(
            ellipse.x - ellipse.rx,
            ellipse.y - ellipse.ry,
            ellipse.x + ellipse.rx,
            ellipse.y + ellipse.ry,
        )
        if self.get_current_value("fill") != momapy.drawing.NoneValue:
            skia_paint = self._make_fill_paint()
            self.canvas.drawOval(oval=skia_rect, paint=skia_paint)
        if self.get_current_value("stroke") != momapy.drawing.NoneValue:
            skia_paint = self._make_stroke_paint()
            self.canvas.drawOval(oval=skia_rect, paint=skia_paint)

    def _render_rectangle(self, rectangle):
        skia_rect = skia.Rect(
            rectangle.x,
            rectangle.y,
            rectangle.x + rectangle.width,
            rectangle.y + rectangle.height,
        )
        if self.get_current_value("fill") != momapy.drawing.NoneValue:
            skia_paint = self._make_fill_paint()
            self.canvas.drawRoundRect(
                rect=skia_rect,
                rx=rectangle.rx,
                ry=rectangle.ry,
                paint=skia_paint,
            )
        if self.get_current_value("stroke") != momapy.drawing.NoneValue:
            skia_paint = self._make_stroke_paint()
            self.canvas.drawRoundRect(
                rect=skia_rect,
                rx=rectangle.rx,
                ry=rectangle.ry,
                paint=skia_paint,
            )

    def _add_move_to(self, skia_path, move_to):
        skia_path.moveTo(move_to.x, move_to.y)

    def _add_line_to(self, skia_path, line_to):
        skia_path.lineTo(line_to.x, line_to.y)

    def _add_curve_to(self, skia_path, curve_to):
        skia_path.cubicTo(
            curve_to.control_point1.x,
            curve_to.control_point1.y,
            curve_to.control_point2.x,
            curve_to.control_point2.y,
            curve_to.x,
            curve_to.y,
        )

    def _add_quadratic_curve_to(self, skia_path, quadratic_curve_to):
        skia_current_point = skia_path.getPoint(skia_path.countPoints() - 1)
        current_point = momapy.geometry.Point(
            skia_current_point.fX, skia_current_point.fY
        )
        curve_to = quadratic_curve_to.to_curve_to(current_point)
        self._add_curve_to(skia_path, curve_to)

    def _add_close_path(self, skia_path, close_path):
        skia_path.close()

    def _add_elliptical_arc(self, skia_path, elliptical_arc):
        if elliptical_arc.arc_flag == 0:
            skia_arc_flag = skia.Path.ArcSize.kSmall_ArcSize
        else:
            skia_arc_flag = skia.Path.ArcSize.kLarge_ArcSize
        if elliptical_arc.sweep_flag == 1:
            skia_sweep_flag = skia.PathDirection.kCW
        else:
            skia_sweep_flag = skia.PathDirection.kCCW
        skia_path.arcTo(
            rx=elliptical_arc.rx,
            ry=elliptical_arc.ry,
            xAxisRotate=elliptical_arc.x_axis_rotation,
            largeArc=skia_arc_flag,
            sweep=skia_sweep_flag,
            x=elliptical_arc.x,
            y=elliptical_arc.y,
        )

    def _add_translation(self, translation):
        self.canvas.translate(dx=translation.tx, dy=translation.ty)

    def _add_rotation(self, rotation):
        angle = math.degrees(rotation.angle)
        if rotation.point is not None:
            self.canvas.rotate(degrees=angle, px=rotation.point.x, py=rotation.point.y)
        else:
            self.canvas.rotate(degrees=angle)

    def _add_scaling(self, scaling):
        self.canvas.scale(sx=scaling.sx, sy=scaling.sy)

    def _add_matrix_transformation(self, matrix_transformation):
        m = skia.Matrix.MakeAll(
            scaleX=matrix_transformation.m[0][0],
            skewX=matrix_transformation.m[0][1],
            transX=matrix_transformation.m[0][2],
            skewY=matrix_transformation.m[1][0],
            scaleY=matrix_transformation.m[1][1],
            transY=matrix_transformation.m[1][2],
            pers0=matrix_transformation.m[2][0],
            pers1=matrix_transformation.m[2][1],
            pers2=matrix_transformation.m[2][2],
        )
        self.canvas.concat(m)


momapy.rendering.core.register_renderer("skia", SkiaRenderer)



"""Classes for SBML maps"""

import dataclasses
import typing
import enum

import momapy.core


class BiomodelQualifier(enum.Enum):
    """Abstract class for http://biomodels.net/ qualifiers"""

    pass


class BQModel(BiomodelQualifier):
    """Class for http://biomodels.net/model-qualifiers/ qualifiers"""

    HAS_INSTANCE = "hasInstance"
    IS = "is"
    IS_DERIVED_FROM = "isDerivedFrom"
    IS_DESCRIBED_BY = "isDescribedBy"
    IS_INSTANCE_OF = "isInstanceOf"


class BQBiol(BiomodelQualifier):
    """Class for http://biomodels.net/biology-qualifiers/ qualifiers"""

    ENCODES = "encodes"
    HAS_PART = "hasPart"
    HAS_PROPERTY = "hasProperty"
    HAS_VERSION = "hasVersion"
    IS = "is"
    IS_DESCRIBED_BY = "isDescribedBy"
    IS_ENCODED_BY = "isEncodedBy"
    IS_HOMOLOG_TO = "isHomologTo"
    IS_PART_OF = "isPartOf"
    IS_PROPERTY_OF = "isPropertyOf"
    IS_VERSION_OF = "isVersionOf"
    OCCURS_IN = "occursIn"
    HAS_TAXON = "hasTaxon"


@dataclasses.dataclass(frozen=True, kw_only=True)
class RDFAnnotation(momapy.core.ModelElement):
    qualifier: BiomodelQualifier
    resources: frozenset[str] = dataclasses.field(default_factory=frozenset)


# to be defined
@dataclasses.dataclass(frozen=True, kw_only=True)
class SBOTerm(momapy.core.ModelElement):
    pass


# abstract
@dataclasses.dataclass(frozen=True, kw_only=True)
class SBase(momapy.core.ModelElement):
    name: str | None = None
    sbo_term: SBOTerm | None = None
    metaid: str | None = dataclasses.field(
        default=None, compare=False, hash=False
    )


@dataclasses.dataclass(frozen=True, kw_only=True)
class Compartment(SBase):
    outside: typing.Optional[
        typing.ForwardRef("Compartment", module="momapy.sbml.core")
    ] = None


@dataclasses.dataclass(frozen=True, kw_only=True)
class Species(SBase):
    compartment: Compartment | None = None


@dataclasses.dataclass(frozen=True, kw_only=True)
class SimpleSpeciesReference(SBase):
    referred_species: Species


@dataclasses.dataclass(frozen=True, kw_only=True)
class ModifierSpeciesReference(SimpleSpeciesReference):
    pass


@dataclasses.dataclass(frozen=True, kw_only=True)
class SpeciesReference(SimpleSpeciesReference):
    stoichiometry: int | None = None


@dataclasses.dataclass(frozen=True, kw_only=True)
class Reaction(SBase):
    reversible: bool
    compartment: Compartment | None = None
    reactants: frozenset[SpeciesReference] = dataclasses.field(
        default_factory=frozenset
    )
    products: frozenset[SpeciesReference] = dataclasses.field(
        default_factory=frozenset
    )
    modifiers: frozenset[ModifierSpeciesReference] = dataclasses.field(
        default_factory=frozenset
    )


@dataclasses.dataclass(frozen=True, kw_only=True)
class Model(SBase, momapy.core.Model):
    compartments: frozenset[Compartment] = dataclasses.field(
        default_factory=frozenset
    )
    species: frozenset[Species] = dataclasses.field(default_factory=frozenset)
    reactions: frozenset[Reaction] = dataclasses.field(
        default_factory=frozenset
    )


@dataclasses.dataclass(frozen=True, kw_only=True)
class SBML(SBase):
    xmlns: str = "http://www.sbml.org/sbml/level3/version2/core"
    level: int = 3
    version: int = 2
    model: Model | None = None



"""Functions for positioning layout elements and related objects relatively to other objects"""

import collections.abc

import momapy.core
import momapy.geometry
import momapy.builder


def right_of(
    obj: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.LayoutElement
        | momapy.core.LayoutElementBuilder
    ),
    distance: float,
) -> momapy.geometry.Point:
    """Compute and return the point right of the given object at a given distance"""
    if momapy.builder.isinstance_or_builder(obj, momapy.geometry.Point):
        source_point = obj
    elif momapy.builder.isinstance_or_builder(obj, momapy.core.Node):
        source_point = obj.east()
    else:
        raise TypeError
    return source_point + (distance, 0)


def left_of(
    obj: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.LayoutElement
        | momapy.core.LayoutElementBuilder
    ),
    distance: float,
) -> momapy.geometry.Point:
    """Compute and return the point left of the given object at a given distance"""
    if momapy.builder.isinstance_or_builder(obj, momapy.geometry.Point):
        source_point = obj
    elif momapy.builder.isinstance_or_builder(obj, momapy.core.Node):
        source_point = obj.west()
    else:
        raise TypeError
    return source_point - (distance, 0)


def above_of(
    obj: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.LayoutElement
        | momapy.core.LayoutElementBuilder
    ),
    distance: float,
) -> momapy.geometry.Point:
    """Compute and return the point above of the given object at a given distance"""
    if momapy.builder.isinstance_or_builder(obj, momapy.geometry.Point):
        source_point = obj
    elif momapy.builder.isinstance_or_builder(obj, momapy.core.Node):
        source_point = obj.north()
    else:
        raise TypeError
    return source_point - (0, distance)


def below_of(
    obj: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.LayoutElement
        | momapy.core.LayoutElementBuilder
    ),
    distance: float,
) -> momapy.geometry.Point:
    """Compute and return the point below of the given object at a given distance"""
    if momapy.builder.isinstance_or_builder(obj, momapy.geometry.Point):
        source_point = obj
    elif momapy.builder.isinstance_or_builder(obj, momapy.core.Node):
        source_point = obj.south()
    else:
        raise TypeError
    return source_point + (0, distance)


def above_left_of(
    obj: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.LayoutElement
        | momapy.core.LayoutElementBuilder
    ),
    distance1: float,
    distance2: float | None = None,
) -> momapy.geometry.Point:
    """Compute and return the point above left of the given object at a given distance"""
    if distance2 is None:
        distance2 = distance1
    if momapy.builder.isinstance_or_builder(obj, momapy.geometry.Point):
        source_point = obj
    elif momapy.builder.isinstance_or_builder(obj, momapy.core.Node):
        source_point = obj.north_west()
    else:
        raise TypeError
    return source_point - (distance2, distance1)


def above_right_of(
    obj: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.LayoutElement
        | momapy.core.LayoutElementBuilder
    ),
    distance1: float,
    distance2: float | None = None,
) -> momapy.geometry.Point:
    """Compute and return the point above right of the given object at a given distance"""
    if distance2 is None:
        distance2 = distance1
    if momapy.builder.isinstance_or_builder(obj, momapy.geometry.Point):
        source_point = obj
    elif momapy.builder.isinstance_or_builder(obj, momapy.core.Node):
        source_point = obj.north_east()
    else:
        raise TypeError
    return source_point + (distance2, -distance1)


def below_left_of(
    obj: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.LayoutElement
        | momapy.core.LayoutElementBuilder
    ),
    distance1: float,
    distance2: float | None = None,
) -> momapy.geometry.Point:
    """Compute and return the point below left of the given object at a given distance"""
    if distance2 is None:
        distance2 = distance1
    if momapy.builder.isinstance_or_builder(obj, momapy.geometry.Point):
        source_point = obj
    elif momapy.builder.isinstance_or_builder(obj, momapy.core.Node):
        source_point = obj.south_west()
    else:
        raise TypeError
    return source_point + (-distance2, distance1)


def below_right_of(
    obj: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.LayoutElement
        | momapy.core.LayoutElementBuilder
    ),
    distance1: float,
    distance2: float | None = None,
) -> momapy.geometry.Point:
    """Compute and return the point below right of the given object at a given distance"""
    if distance2 is None:
        distance2 = distance1
    if momapy.builder.isinstance_or_builder(obj, momapy.geometry.Point):
        source_point = obj
    elif momapy.builder.isinstance_or_builder(obj, momapy.core.Node):
        source_point = obj.south_east()
    else:
        raise TypeError
    return source_point + (distance2, distance1)


def fit(
    elements: collections.abc.Collection[
        momapy.core.LayoutElement
        | momapy.core.LayoutElementBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.geometry.Point
        | momapy.geometry.PointBuilder
    ],
    xsep: float = 0,
    ysep: float = 0,
) -> momapy.geometry.Bbox:
    """Compute and return the bounding box fitting a collection of objects, with given margins"""
    if not elements:
        raise ValueError("elements must contain at least one element")
    points = []
    for element in elements:
        if momapy.builder.isinstance_or_builder(element, momapy.geometry.Point):
            points.append(element)
        elif momapy.builder.isinstance_or_builder(element, momapy.geometry.Bbox):
            points.append(element.north_west())
            points.append(element.south_east())
        elif momapy.builder.isinstance_or_builder(element, momapy.core.LayoutElement):
            bbox = element.bbox()
            points.append(bbox.north_west())
            points.append(bbox.south_east())
        else:
            raise TypeError(f"{type(element)} not supported")
    point = points[0]
    max_x = point.x
    max_y = point.y
    min_x = point.x
    min_y = point.y
    for point in points[1:]:
        if point.x > max_x:
            max_x = point.x
        elif point.x < min_x:
            min_x = point.x
        if point.y > max_y:
            max_y = point.y
        elif point.y < min_y:
            min_y = point.y
    max_x += xsep
    min_x -= xsep
    max_y += ysep
    min_y -= ysep
    width = max_x - min_x
    height = max_y - min_y
    bbox = momapy.geometry.Bbox(
        momapy.geometry.Point(min_x + width / 2, min_y + height / 2),
        width,
        height,
    )
    return bbox


def mid_of(
    obj1: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.Node
        | momapy.core.NodeBuilder
    ),
    obj2: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.Node
        | momapy.core.NodeBuilder
    ),
):
    if momapy.builder.isinstance_or_builder(
        obj1, (momapy.geometry.Bbox, momapy.core.Node)
    ):
        obj1 = obj1.center()
    if momapy.builder.isinstance_or_builder(
        obj2, (momapy.geometry.Bbox, momapy.core.Node)
    ):
        obj2 = obj2.center()
    segment = momapy.geometry.Segment(obj1, obj2)
    return segment.get_position_at_fraction(0.5)


def cross_vh_of(
    obj1: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.Node
        | momapy.core.NodeBuilder
    ),
    obj2: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.Node
        | momapy.core.NodeBuilder
    ),
):
    if momapy.builder.isinstance_or_builder(
        obj1, (momapy.geometry.Bbox, momapy.core.Node)
    ):
        obj1 = obj1.center()
    if momapy.builder.isinstance_or_builder(
        obj2, (momapy.geometry.Bbox, momapy.core.Node)
    ):
        obj2 = obj2.center()
    x = obj1.x
    y = obj2.y
    return momapy.geometry.Point(x, y)


def cross_hv_of(
    obj1: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.Node
        | momapy.core.NodeBuilder
    ),
    obj2: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.Node
        | momapy.core.NodeBuilder
    ),
):
    if momapy.builder.isinstance_or_builder(
        obj1, (momapy.geometry.Bbox, momapy.core.Node)
    ):
        obj1 = obj1.center()
    if momapy.builder.isinstance_or_builder(
        obj2, (momapy.geometry.Bbox, momapy.core.Node)
    ):
        obj2 = obj2.center()
    y = obj1.y
    x = obj2.x
    return momapy.geometry.Point(x, y)


def fraction_of(
    arc_layout_element: (momapy.core.SingleHeadedArc | momapy.core.DoubleHeadedArc),
    fraction: float,
) -> tuple[momapy.geometry.Point, float]:
    """Return the position on an arc at a given fraction and the angle formed of the tangant of the arc at that position and the horizontal"""
    position, angle = arc_layout_element.fraction(fraction)
    return position, angle


def set_position(
    obj: momapy.core.NodeBuilder | momapy.geometry.BboxBuilder,
    position: momapy.geometry.Point | momapy.geometry.PointBuilder,
    anchor: str | None = None,
):
    """Set the position of a given builder object"""
    obj.position = position
    if anchor is not None:
        p = getattr(obj, anchor)()
        obj.position += obj.position - p


def set_right_of(
    obj1: momapy.core.NodeBuilder | momapy.geometry.BboxBuilder,
    obj2: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.LayoutElement
        | momapy.core.LayoutElementBuilder
    ),
    distance: float,
    anchor: str | None = None,
):
    """Set the position of a given builder object right of another given object at a given distance"""
    position = right_of(obj2, distance)
    set_position(obj1, position, anchor)


def set_left_of(
    obj1: momapy.core.NodeBuilder | momapy.geometry.BboxBuilder,
    obj2: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.LayoutElement
        | momapy.core.LayoutElementBuilder
    ),
    distance: float,
    anchor: str | None = None,
):
    """Set the position of a given builder object left of another given object at a given distance"""
    position = left_of(obj2, distance)
    set_position(obj1, position, anchor)


def set_above_of(
    obj1: momapy.core.NodeBuilder | momapy.geometry.BboxBuilder,
    obj2: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.LayoutElement
        | momapy.core.LayoutElementBuilder
    ),
    distance: float,
    anchor: str | None = None,
):
    """Set the position of a given builder object above of another given object at a given distance"""
    position = above_of(obj2, distance)
    set_position(obj1, position, anchor)


def set_below_of(
    obj1: momapy.core.NodeBuilder | momapy.geometry.BboxBuilder,
    obj2: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.LayoutElement
        | momapy.core.LayoutElementBuilder
    ),
    distance: float,
    anchor: str | None = None,
):
    """Set the position of a given builder object below of another given object at a given distance"""
    position = below_of(obj2, distance)
    set_position(obj1, position, anchor)


def set_above_left_of(
    obj1: momapy.core.NodeBuilder | momapy.geometry.BboxBuilder,
    obj2: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.LayoutElement
        | momapy.core.LayoutElementBuilder
    ),
    distance1: float,
    distance2: float | None = None,
    anchor: str | None = None,
):
    """Set the position of a given builder object above left of another given object at a given distance"""
    position = above_left_of(obj2, distance1, distance2)
    set_position(obj1, position, anchor)


def set_above_right_of(
    obj1: momapy.core.NodeBuilder | momapy.geometry.BboxBuilder,
    obj2: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.LayoutElement
        | momapy.core.LayoutElementBuilder
    ),
    distance1: float,
    distance2: float | None = None,
    anchor: str | None = None,
):
    """Set the position of a given builder object above right of another given object at a given distance"""
    position = above_right_of(obj2, distance1, distance2)
    set_position(obj1, position, anchor)


def set_below_left_of(
    obj1: momapy.core.NodeBuilder | momapy.geometry.BboxBuilder,
    obj2: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.LayoutElement
        | momapy.core.LayoutElementBuilder
    ),
    distance1: float,
    distance2: float | None = None,
    anchor: str | None = None,
):
    """Set the position of a given builder object below left of another given object at a given distance"""
    position = below_left_of(obj2, distance1, distance2)
    set_position(obj1, position, anchor)


def set_below_right_of(
    obj1: momapy.core.NodeBuilder | momapy.geometry.BboxBuilder,
    obj2: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.LayoutElement
        | momapy.core.LayoutElementBuilder
    ),
    distance1: float,
    distance2: float | None = None,
    anchor: str | None = None,
):
    """Set the position of a given builder object below right of another given object at a given distance"""
    position = below_right_of(obj2, distance1, distance2)
    set_position(obj1, position, anchor)


def set_fit(
    obj: momapy.core.NodeBuilder | momapy.geometry.BboxBuilder,
    elements: collections.abc.Collection[
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.LayoutElement
        | momapy.core.LayoutElementBuilder
    ],
    xsep: float = 0,
    ysep: float = 0,
    anchor: str | None = None,
):
    """Set the position, width and height of a given builder object to fit a given collection of other objects"""
    bbox = fit(elements, xsep, ysep)
    obj.width = bbox.width
    obj.height = bbox.height
    set_position(obj, bbox.position, anchor)


def set_fraction_of(
    obj: momapy.core.NodeBuilder,
    arc_layout_element: (momapy.core.SingleHeadedArc | momapy.core.DoubleHeadedArc),
    fraction: float,
    anchor: str | None = None,
):
    """Set the position and rotation of a given builder object to lay at a given fraction of a given arc"""
    position, angle = fraction_of(arc_layout_element, fraction)
    rotation = momapy.geometry.Rotation(angle, position)
    set_position(obj, position, anchor)
    obj.transform = momapy.core.TupleBuilder([rotation])


def set_mid_of(
    obj1: momapy.core.NodeBuilder | momapy.geometry.BboxBuilder,
    obj2: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.Node
        | momapy.core.NodeBuilder
    ),
    obj3: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.Node
        | momapy.core.NodeBuilder
    ),
    anchor: str | None = None,
):
    position = mid_of(obj2, obj3)
    set_position(obj1, position, anchor)


def set_cross_hv_of(
    obj1: momapy.core.NodeBuilder | momapy.geometry.BboxBuilder,
    obj2: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.Node
        | momapy.core.NodeBuilder
    ),
    obj3: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.Node
        | momapy.core.NodeBuilder
    ),
    anchor: str | None = None,
):
    position = cross_hv_of(obj2, obj3)
    set_position(obj1, position, anchor)


def set_cross_vh_of(
    obj1: momapy.core.NodeBuilder | momapy.geometry.BboxBuilder,
    obj2: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.Node
        | momapy.core.NodeBuilder
    ),
    obj3: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.Node
        | momapy.core.NodeBuilder
    ),
    anchor: str | None = None,
):
    position = cross_vh_of(obj2, obj3)
    set_position(obj1, position, anchor)



from dataclasses import dataclass
import tempfile
import ffmpeg

import momapy.core
import momapy.rendering


@dataclass
class Animator(object):
    layout: momapy.core.Layout
    fps: int = 60

    def __post_init__(self):
        self._initialize()

    def _initialize(self):
        flimages = tempfile.mkstemp()
        self._flimages = (open(flimages[1], "w"), flimages[1])
        self._n_images = 0

    def frames(self, n_frames: int):
        fimage = tempfile.mkstemp()
        self._n_images += 1
        momapy.rendering.render_layout(self.layout, fimage[1], format_="png")
        for i in range(n_frames):
            self._flimages[0].write(f"file '{fimage[1]}\n")

    def mseconds(self, n_mseconds: float):
        n_frames = round(n_mseconds / 1000 * self.fps)
        self.frames(n_frames)

    def build(self, output_file, vcodec="libx264"):
        self._flimages[0].close()
        ffmpeg.input(
            self._flimages[1], r=str(self.fps), f="concat", safe="0"
        ).output(output_file, vcodec=vcodec).run(
            quiet=True, overwrite_output=True
        )



"""Classes and functions for styling layout elements using style sheets"""

import abc
import collections.abc
import dataclasses
import pyparsing
import copy


import momapy.coloring
import momapy.drawing
import momapy.core


class StyleCollection(dict):
    """Class for style collections"""

    pass


class StyleSheet(dict):
    """Class for style sheets"""

    def __or__(self, other):
        d = copy.deepcopy(self)
        for key, value in other.items():
            if key in d:
                d[key] |= value
            else:
                d[key] = value
        return StyleSheet(d)

    def __ior__(self, other):
        return self.__or__(other)

    @classmethod
    def from_file(cls, file_path: str) -> "StyleSheet":
        """Read and return a style sheet from a file"""
        style_sheet = _css_document.parse_file(file_path, parse_all=True)[0]
        return style_sheet

    @classmethod
    def from_string(cls, s: str) -> "StyleSheet":
        """Read and return a style sheet from a string"""
        style_sheet = _css_document.parse_string(s, parse_all=True)[0]
        return style_sheet

    @classmethod
    def from_files(cls, file_paths: collections.abc.Collection[str]) -> "StyleSheet":
        """Read and return a style sheet from a collection of files"""
        style_sheets = []
        for file_path in file_paths:
            style_sheet = StyleSheet.from_file(file_path)
            style_sheets.append(style_sheet)
        style_sheet = combine_style_sheets(style_sheets)
        return style_sheet


def combine_style_sheets(
    style_sheets: collections.abc.Collection[StyleSheet],
) -> StyleSheet:
    """Merge a collection of style sheets into a unique style sheet and return it"""
    if not style_sheets:
        return None
    output_style_sheet = style_sheets[0]
    for style_sheet in style_sheets[1:]:
        output_style_sheet |= style_sheet
    return output_style_sheet


def apply_style_collection(
    layout_element: (momapy.core.LayoutElement | momapy.core.LayoutElementBuilder),
    style_collection: StyleCollection,
    strict: bool = True,
) -> momapy.core.LayoutElement | momapy.core.LayoutElementBuilder:
    """Apply a style collection to a layout element"""
    if not isinstance(layout_element, momapy.builder.Builder):
        layout_element = momapy.builder.builder_from_object(layout_element)
        is_builder = False
    else:
        is_builder = True
    for attribute, value in style_collection.items():
        if hasattr(layout_element, attribute):
            setattr(layout_element, attribute, value)
        else:
            if strict:
                raise AttributeError(
                    f"{type(layout_element)} object has no attribute '{attribute}'"
                )
    if is_builder:
        return layout_element
    return momapy.builder.object_from_builder(layout_element)


def apply_style_sheet(
    map_or_layout_element: (
        momapy.core.Map
        | momapy.core.LayoutElement
        | momapy.core.MapBuilder
        | momapy.core.LayoutElementBuilder
    ),
    style_sheet: StyleSheet,
    strict: bool = True,
    ancestors: collections.abc.Collection[
        momapy.core.LayoutElement | momapy.core.LayoutElementBuilder
    ] = None,
) -> (
    momapy.core.Map
    | momapy.core.LayoutElement
    | momapy.core.LayoutElementBuilder
    | momapy.core.MapBuilder
):
    """Apply a style sheet to a layout element or (layout of) a map"""
    if not isinstance(map_or_layout_element, momapy.builder.Builder):
        map_or_layout_element = momapy.builder.builder_from_object(
            map_or_layout_element
        )
        is_builder = False
    else:
        is_builder = True
    if isinstance(map_or_layout_element, momapy.core.MapBuilder):
        layout_element = map_or_layout_element.layout
    else:
        layout_element = map_or_layout_element
    if style_sheet is not None:
        if ancestors is None:
            ancestors = []
        for selector, style_collection in style_sheet.items():
            if selector.select(layout_element, ancestors):
                apply_style_collection(
                    layout_element=layout_element,
                    style_collection=style_collection,
                    strict=strict,
                )
        ancestors = ancestors + [layout_element]
        for child in layout_element.children():
            apply_style_sheet(
                map_or_layout_element=child,
                style_sheet=style_sheet,
                strict=strict,
                ancestors=ancestors,
            )
    if is_builder:
        return map_or_layout_element
    return momapy.builder.object_from_builder(map_or_layout_element)


@dataclasses.dataclass(frozen=True)
class Selector(object):
    """Abstract class for selectors"""

    @abc.abstractmethod
    def select(
        self,
        obj: momapy.core.LayoutElement | momapy.core.LayoutElementBuilder,
        ancestors: collections.abc.Collection[
            momapy.core.LayoutElement | momapy.core.LayoutElementBuilder
        ],
    ) -> bool:
        """Return `true` if the given layout element satisfies the selector, and `false` otherwise"""
        pass


@dataclasses.dataclass(frozen=True)
class TypeSelector(Selector):
    """Class for type selectors"""

    class_name: str = dataclasses.field(
        metadata={"description": "The name of the class"}
    )

    def select(
        self,
        obj: momapy.core.LayoutElement | momapy.core.LayoutElementBuilder,
        ancestors: collections.abc.Collection[
            momapy.core.LayoutElement | momapy.core.LayoutElementBuilder
        ],
    ):
        """Return `true` if the given layout element satisfies the selector, and `false` otherwise"""
        obj_cls_name = type(obj).__name__
        return (
            obj_cls_name == self.class_name
            or obj_cls_name == f"{self.class_name}Builder"
        )


@dataclasses.dataclass(frozen=True)
class ClassSelector(Selector):
    """Class for class selectors"""

    class_name: str = dataclasses.field(
        metadata={"description": "The name of the class"}
    )

    def select(
        self,
        obj: momapy.core.LayoutElement | momapy.core.LayoutElementBuilder,
        ancestors: collections.abc.Collection[
            momapy.core.LayoutElement | momapy.core.LayoutElementBuilder
        ],
    ):
        """Return `true` if the given layout element satisfies the selector, and `false` otherwise"""
        for cls in type(obj).__mro__:
            cls_name = cls.__name__
            # print(cls_name, f"{self.class_name}Builder")
            if cls_name == self.class_name or cls_name == f"{self.class_name}Builder":
                return True
        return False


@dataclasses.dataclass(frozen=True)
class IdSelector(Selector):
    """Class for id selectors"""

    id_: str = dataclasses.field(metadata={"description": "The id"})

    def select(
        self,
        obj: momapy.core.LayoutElement | momapy.core.LayoutElementBuilder,
        ancestors: collections.abc.Collection[
            momapy.core.LayoutElement | momapy.core.LayoutElementBuilder
        ],
    ):
        """Return `true` if the given layout element satisfies the selector, and `false` otherwise"""
        return hasattr(obj, "id_") and obj.id_ == self.id_


@dataclasses.dataclass(frozen=True)
class ChildSelector(Selector):
    """Class for child selectors"""

    parent_selector: Selector = dataclasses.field(
        metadata={"description": "The parent selector"}
    )
    child_selector: Selector = dataclasses.field(
        metadata={"description": "The child selector"}
    )

    def select(
        self,
        obj: momapy.core.LayoutElement | momapy.core.LayoutElementBuilder,
        ancestors: collections.abc.Collection[
            momapy.core.LayoutElement | momapy.core.LayoutElementBuilder
        ],
    ):
        """Return `true` if the given layout element satisfies the selector, and `false` otherwise"""
        if not ancestors:
            return False
        return self.child_selector.select(
            obj, ancestors
        ) and self.parent_selector.select(ancestors[-1], ancestors[:-1])


@dataclasses.dataclass(frozen=True)
class DescendantSelector(Selector):
    """Class for descendant selectors"""

    ancestor_selector: Selector = dataclasses.field(
        metadata={"description": "The ancestor selector"}
    )
    descendant_selector: Selector = dataclasses.field(
        metadata={"description": "The descendant selector"}
    )

    def select(
        self,
        obj: momapy.core.LayoutElement | momapy.core.LayoutElementBuilder,
        ancestors: collections.abc.Collection[
            momapy.core.LayoutElement | momapy.core.LayoutElementBuilder
        ],
    ):
        """Return `true` if the given layout element satisfies the selector, and `false` otherwise"""
        if not ancestors:
            return False
        return self.descendant_selector.select(obj, ancestors) and any(
            [
                self.ancestor_selector.select(ancestor, ancestors[:i])
                for i, ancestor in enumerate(ancestors)
            ]
        )


@dataclasses.dataclass(frozen=True)
class OrSelector(Selector):
    """Class for or selectors"""

    selectors: tuple[Selector, ...] = dataclasses.field(
        metadata={"description": "The tuple of disjunct selectors"}
    )

    def select(
        self,
        obj: momapy.core.LayoutElement | momapy.core.LayoutElementBuilder,
        ancestors: collections.abc.Collection[
            momapy.core.LayoutElement | momapy.core.LayoutElementBuilder
        ],
    ):
        """Return `true` if the given layout element satisfies the selector, and `false` otherwise"""
        return any([selector.select(obj, ancestors) for selector in self.selectors])


@dataclasses.dataclass(frozen=True)
class CompoundSelector(Selector):
    """Class for compound selectors"""

    selectors: tuple[Selector, ...] = dataclasses.field(
        metadata={"description": "The tuple of conjunct selectors"}
    )

    def select(
        self,
        obj: momapy.core.LayoutElement | momapy.core.LayoutElementBuilder,
        ancestors: collections.abc.Collection[
            momapy.core.LayoutElement | momapy.core.LayoutElementBuilder
        ],
    ):
        """Return `true` if the given layout element satisfies the selector, and `false` otherwise"""
        return all([selector.select(obj, ancestors) for selector in self.selectors])


@dataclasses.dataclass(frozen=True)
class NotSelector(Selector):
    """Class for not selectors"""

    selectors: tuple[Selector, ...] = dataclasses.field(
        metadata={"description": "The tuple of negated conjunct selectors"}
    )

    def select(
        self,
        obj: momapy.core.LayoutElement | momapy.core.LayoutElementBuilder,
        ancestors: collections.abc.Collection[
            momapy.core.LayoutElement | momapy.core.LayoutElementBuilder
        ],
    ):
        """Return `true` if the given layout element satisfies the selector, and `false` otherwise"""
        return not any([selector.select(obj, ancestors) for selector in self.selectors])


_css_import_keyword = pyparsing.Literal("@import")
_css_unset_value = pyparsing.Literal("unset")
_css_none_value = pyparsing.Literal("none")
_css_float_value = pyparsing.Combine(
    pyparsing.Word(pyparsing.nums)
    + pyparsing.Literal(".")
    + pyparsing.Word(pyparsing.nums)
)
_css_string_value = pyparsing.quoted_string
_css_color_name_value = pyparsing.Word(pyparsing.alphas + "_")
_css_color_value = _css_color_name_value
_css_int_value = pyparsing.Word(pyparsing.nums)
_css_drop_shadow_filter_value = (
    pyparsing.Literal("drop-shadow(")
    + _css_float_value
    + pyparsing.Literal(",")
    + _css_float_value
    + pyparsing.Literal(",")
    + _css_float_value
    + pyparsing.Literal(",")
    + _css_float_value
    + pyparsing.Literal(",")
    + _css_color_value
    + pyparsing.Literal(")")
)
_css_filter_value = _css_drop_shadow_filter_value
_css_simple_value = (
    _css_drop_shadow_filter_value
    | _css_unset_value
    | _css_none_value
    | _css_float_value
    | _css_string_value
    | _css_color_value
    | _css_int_value
)
_css_list_value = pyparsing.Group(
    pyparsing.delimited_list(_css_simple_value, ",", min=2)
)
_css_attribute_value = _css_list_value | _css_simple_value
_css_attribute_name = pyparsing.Word(
    pyparsing.alphas + "_", pyparsing.alphanums + "_" + "-"
)


_css_import_statement = _css_import_keyword + _css_string_value + pyparsing.Literal(";")
_css_style = (
    _css_attribute_name
    + pyparsing.Literal(":")
    + _css_attribute_value
    + pyparsing.Literal(";")
)
_css_style_collection = (
    pyparsing.Literal("{")
    + pyparsing.Group(_css_style[1, ...])
    + pyparsing.Literal("}")
)
_css_id = pyparsing.Word(pyparsing.printables, exclude_chars=",")
_css_id_selector = pyparsing.Literal("#") + _css_id
_css_class_name = pyparsing.Word(pyparsing.alphas + "_", pyparsing.alphanums + "_")
_css_type_selector = _css_class_name.copy()
_css_class_selector = pyparsing.Literal(".") + _css_class_name
_css_elementary_selector = _css_class_selector | _css_type_selector | _css_id_selector
_css_child_selector = (
    _css_elementary_selector + pyparsing.Literal(">") + _css_elementary_selector
)
_css_descendant_selector = (
    _css_elementary_selector
    + pyparsing.OneOrMore(pyparsing.White())
    + _css_elementary_selector
)
_css_or_selector = pyparsing.Group(
    pyparsing.delimited_list(_css_elementary_selector, ",", min=2)
)
_css_selector = (
    _css_child_selector
    | _css_descendant_selector
    | _css_or_selector
    | _css_elementary_selector
)
_css_rule = _css_selector + _css_style_collection
_css_style_sheet = pyparsing.Group(_css_rule[1, ...])
_css_document = _css_import_statement[...] + _css_style_sheet[...]


@_css_unset_value.set_parse_action
def _resolve_css_unset_value(results):
    return results[0]


@_css_none_value.set_parse_action
def _resolve_css_none_value(results):
    return momapy.drawing.NoneValue


@_css_float_value.set_parse_action
def _resolve_css_float_value(results):
    return float(results[0])


@_css_string_value.set_parse_action
def _resolve_css_string_value(results):
    return str(results[0][1:-1])


@_css_int_value.set_parse_action
def _resolve_css_int_value(results):
    return int(results[0])


@_css_color_name_value.set_parse_action
def _resolve_css_color_name_value(results):
    if not momapy.coloring.has_color(results[0]):
        raise ValueError(f"{results[0]} is not a valid color name")
    return getattr(momapy.coloring, results[0])


@_css_drop_shadow_filter_value.set_parse_action
def _resolve_css_drop_shadow_filter_value(results):
    filter_effect = momapy.builder.get_or_make_builder_cls(
        momapy.drawing.DropShadowEffect
    )(
        dx=results[1],
        dy=results[3],
        std_deviation=results[5],
        flood_opacity=results[7],
        flood_color=results[9],
    )
    filter = momapy.builder.get_or_make_builder_cls(momapy.drawing.Filter)(
        effects=momapy.core.TupleBuilder([filter_effect])
    )
    return filter


# Issue: the function cannot return None (pyparsing bug?) otherwise it simply
# does not apply the function
@_css_simple_value.set_parse_action
def _resolve_css_simple_value(results):
    return results[0]


@_css_list_value.set_parse_action
def _resolve_css_list_value(results):
    return [momapy.core.TupleBuilder(results[0])]


@_css_attribute_value.set_parse_action
def _resolve_css_attribute_value(results):
    # see above
    if results[0] == "unset":
        results[0] = None
    return results


@_css_import_statement.set_parse_action
def _resolve_css_import_statement(results):
    file_path = results[1]
    style_sheet = StyleSheet.from_file(file_path)
    return style_sheet


@_css_attribute_name.set_parse_action
def _resolve_css_attribute_name(results):
    return results[0].replace("-", "_")


@_css_style.set_parse_action
def _resolve_css_style(results):
    return (
        results[0],
        results[2],
    )


@_css_style_collection.set_parse_action
def _resolve_css_style_collection(results):
    return StyleCollection(dict(list(results[1])))


@_css_id.set_parse_action
def _resolve_css_id(results):
    return results[0]


@_css_id_selector.set_parse_action
def _resolve_id_selector(results):
    return IdSelector(results[1])


@_css_class_name.set_parse_action
def _resolve_css_class_name(results):
    return results[0]


@_css_type_selector.set_parse_action
def _resolve_css_type_selector(results):
    return TypeSelector(results[0])


@_css_class_selector.set_parse_action
def _resolve_css_class_selector(results):
    return ClassSelector(results[1])


@_css_elementary_selector.set_parse_action
def _resolve_css_elementary_selector(results):
    return results[0]


@_css_child_selector.set_parse_action
def _resolve_css_child_selector(results):
    return ChildSelector(results[0], results[2])


@_css_descendant_selector.set_parse_action
def _resolve_css_descendant_selector(results):
    return DescendantSelector(results[0], results[2])


@_css_or_selector.set_parse_action
def _resolve_css_or_selector(results):
    return OrSelector(tuple(results[0]))


@_css_rule.set_parse_action
def _resolve_css_rule(results):
    return (
        results[0],
        results[1],
    )


@_css_style_sheet.set_parse_action
def _resolve_css_style_sheet(results):
    return StyleSheet(dict(list(results[0])))


@_css_document.set_parse_action
def _resolve_css_document(results):
    style_sheets = [style_sheet for style_sheet in results if style_sheet is not None]
    return combine_style_sheets(style_sheets)



"""Subpackage for rendering"""

_registered = False


def _ensure_registered():
    global _registered
    if not _registered:
        import momapy.rendering.core
        import momapy.rendering.svg_native

        momapy.rendering.core.register_renderer(
            "svg-native", momapy.rendering.svg_native.SVGNativeRenderer
        )
        momapy.rendering.core.register_renderer(
            "svg-native-compat", momapy.rendering.svg_native.SVGNativeCompatRenderer
        )



import dataclasses
import typing
import abc

_registered_object_callbacks = {}
_registered_attribute_callbacks = {}


@dataclasses.dataclass(frozen=True)
class Event(abc.ABC):
    obj: typing.Any


@dataclasses.dataclass(frozen=True)
class ChangedEvent(Event):
    attr_name: None | str = None


@dataclasses.dataclass(frozen=True)
class SetEvent(Event):
    attr_name: str


def register_event(obj, event_cls, callback, attr_name=None):
    if attr_name is None:
        if id(obj) not in _registered_object_callbacks:
            _registered_object_callbacks[id(obj)] = {}
        if event_cls not in _registered_object_callbacks[id(obj)]:
            _registered_object_callbacks[id(obj)][event_cls] = []
        _registered_object_callbacks[id(obj)][event_cls].append(callback)
        if (
            event_cls == ChangedEvent or event_cls == SetEvent
        ) and dataclasses.is_dataclass(obj):
            for field_ in dataclasses.fields(obj):
                field_name = field_.name
                register_event(
                    obj,
                    event_cls,
                    lambda event: trigger_event(event_cls(obj)),
                    field_name,
                )
    else:
        if id(obj) not in _registered_attribute_callbacks:
            _registered_attribute_callbacks[id(obj)] = {}
        if attr_name not in _registered_attribute_callbacks[id(obj)]:
            _registered_attribute_callbacks[id(obj)][attr_name] = {}
        if event_cls not in _registered_attribute_callbacks[id(obj)][attr_name]:
            _registered_attribute_callbacks[id(obj)][attr_name][event_cls] = []
        _registered_attribute_callbacks[id(obj)][attr_name][event_cls].append(
            callback
        )
        if event_cls == ChangedEvent:
            on_set(
                obj,
                lambda event: trigger_event(ChangedEvent(obj, attr_name)),
                attr_name,
            )
            on_change(
                getattr(obj, attr_name),
                lambda event: trigger_event(ChangedEvent(obj, attr_name)),
            )


def on_change(obj, callback, attr_name=None):
    register_event(obj, ChangedEvent, callback, attr_name)


def on_set(obj, callback, attr_name=None):
    register_event(obj, SetEvent, callback, attr_name)


def trigger_event(event):
    if event.attr_name is None:
        if (
            id(event.obj) in _registered_object_callbacks
            and type(event) in _registered_object_callbacks[id(event.obj)]
        ):
            callbacks = _registered_object_callbacks[id(event.obj)][type(event)]
        else:
            callbacks = []
    else:
        if (
            id(event.obj) in _registered_attribute_callbacks
            and event.attr_name
            in _registered_attribute_callbacks[id(event.obj)]
            and type(event)
            in _registered_attribute_callbacks[id(event.obj)][event.attr_name]
        ):
            callbacks = _registered_attribute_callbacks[id(event.obj)][
                event.attr_name
            ][type(event)]
        else:
            callbacks = []
    for callback in callbacks:
        callback(event)


@dataclasses.dataclass
class Monitored:
    def __setattr__(self, name, value):
        super().__setattr__(name, value)
        trigger_event(SetEvent(self, name))



"""Tests for momapy.io.core module."""
import pytest
import momapy.io.core


def test_readers_dict_exists():
    """Test that readers dictionary exists."""
    assert isinstance(momapy.io.core.readers, dict)


def test_writers_dict_exists():
    """Test that writers dictionary exists."""
    assert isinstance(momapy.io.core.writers, dict)


def test_register_reader():
    """Test register_reader function."""
    class DummyReader:
        @staticmethod
        def check_file(file_path):
            return False

    # Save original state
    original_readers = momapy.io.core.readers.copy()

    try:
        momapy.io.core.register_reader("test_reader", DummyReader)
        assert "test_reader" in momapy.io.core.readers
        assert momapy.io.core.readers["test_reader"] == DummyReader
    finally:
        # Restore original state
        momapy.io.core.readers.clear()
        momapy.io.core.readers.update(original_readers)


def test_register_writer():
    """Test register_writer function."""
    class DummyWriter:
        pass

    # Save original state
    original_writers = momapy.io.core.writers.copy()

    try:
        momapy.io.core.register_writer("test_writer", DummyWriter)
        assert "test_writer" in momapy.io.core.writers
        assert momapy.io.core.writers["test_writer"] == DummyWriter
    finally:
        # Restore original state
        momapy.io.core.writers.clear()
        momapy.io.core.writers.update(original_writers)


def test_io_result_creation():
    """Test IOResult creation."""
    result = momapy.io.core.IOResult()
    assert result.exceptions == []


def test_reader_result_creation():
    """Test ReaderResult creation."""
    result = momapy.io.core.ReaderResult()
    assert result.exceptions == []
    assert result.obj is None


def test_writer_result_creation():
    """Test WriterResult creation."""
    result = momapy.io.core.WriterResult()
    assert result.exceptions == []
    assert result.obj is None


def test_read_function_exists():
    """Test that read function exists."""
    assert hasattr(momapy.io.core, 'read')
    assert callable(momapy.io.core.read)


def test_write_function_exists():
    """Test that write function exists."""
    assert hasattr(momapy.io.core, 'write')
    assert callable(momapy.io.core.write)



"""Classes and functions for geometry"""

import dataclasses
import typing
import typing_extensions
import abc
import math
import copy

import numpy
import shapely
import shapely.ops
import bezier.curve

import momapy.builder


ROUNDING = 2


@dataclasses.dataclass(frozen=True)
class GeometryObject(abc.ABC):
    """Base class for geometry objects"""

    @abc.abstractmethod
    def to_shapely(self) -> shapely.Geometry:
        """Compute and return a shapely geometry object that reproduces the geometry object"""
        pass


@dataclasses.dataclass(frozen=True)
class Point(GeometryObject):
    """Class for points"""

    x: float
    y: float

    def __post_init__(self):
        object.__setattr__(self, "x", round(self.x, ROUNDING))
        object.__setattr__(self, "y", round(self.y, ROUNDING))

    def __add__(self, xy):
        if momapy.builder.isinstance_or_builder(xy, Point):
            xy = (xy.x, xy.y)
        return Point(self.x + xy[0], self.y + xy[1])

    def __sub__(self, xy):
        if momapy.builder.isinstance_or_builder(xy, Point):
            xy = (xy.x, xy.y)
        return Point(self.x - xy[0], self.y - xy[1])

    def __mul__(self, scalar):
        return Point(self.x * scalar, self.y * scalar)

    def __truediv__(self, scalar):
        return Point(self.x / scalar, self.y / scalar)

    def __iter__(self):
        yield self.x
        yield self.y

    def to_matrix(self) -> numpy.ndarray:
        """Return a `numpy.array` representation of the point"""
        m = numpy.array([[self.x], [self.y], [1]], dtype=float)
        return m

    def to_tuple(self) -> tuple[float, float]:
        """Return a tuple representation of the point"""
        return (
            self.x,
            self.y,
        )

    def get_intersection_with_line(self, line: "Line") -> list["Point"]:
        """Compute and return a list of the intersections of the point and a given line"""
        return get_intersection_of_line_and_point(line, self)

    def get_angle_to_horizontal(self) -> float:
        """Return the angle in radians formed by the line passing through the origin and the point and the horizontal"""
        return get_angle_to_horizontal_of_point(self)

    def transformed(self, transformation: "Transformation") -> "Point":
        """Return a copy of the the point transformed by the given transformation"""
        return transform_point(self, transformation)

    def reversed(self) -> "Point":
        """Return a reversed copy of the point"""
        return reverse_point(self)

    def round(self, ndigits=None):
        return Point(round(self.x, ndigits), round(self.y, ndigits))

    def to_shapely(self) -> shapely.Point:
        """Return a shapely point that reproduces the point"""
        return shapely.Point(self.x, self.y)

    def to_fortranarray(self) -> typing.Any:
        """Return a numpy fortran array representation of the point"""
        return numpy.asfortranarray([[self.x], [self.y]])

    def bbox(self) -> "Bbox":
        """Return the bounding box of the point"""
        return Bbox(copy.deepcopy(self), 0, 0)

    def isnan(self) -> bool:
        """Return `true` if the point has a nan coordinate, `false` otherwise"""
        return math.isnan(self.x) or math.isnan(self.y)

    @classmethod
    def from_shapely(cls, point: shapely.Point) -> typing_extensions.Self:
        """Return a point reproducing a given shapely point"""
        return cls(float(point.x), float(point.y))

    @classmethod
    def from_fortranarray(cls, fortranarray: typing.Any) -> typing_extensions.Self:
        """Return a point from a numpy fortran array representation"""
        return cls(fortranarray[0][0], fortranarray[1][1])

    @classmethod
    def from_tuple(cls, t: tuple[float, float]) -> typing_extensions.Self:
        """Return a point from a tuple representation"""
        return cls(t[0], t[1])


@dataclasses.dataclass(frozen=True)
class Line(GeometryObject):
    """Class for lines"""

    p1: Point
    p2: Point

    def slope(self) -> float:
        """Return the slope of the line"""
        if self.p1.x != self.p2.x:
            return round((self.p2.y - self.p1.y) / (self.p2.x - self.p1.x), ROUNDING)
        return float("NAN")  # infinite slope

    def intercept(self) -> float:
        """Return the intercept of the line"""
        slope = self.slope()
        if not math.isnan(slope):
            return self.p1.y - slope * self.p1.x
        else:
            return float("NAN")

    def get_angle_to_horizontal(self) -> float:
        """Return the angle in radians formed by the line and the horizontal"""
        return get_angle_to_horizontal_of_line(self)

    def is_parallel_to_line(self, line: "Line") -> bool:
        """Return `true` if the line is parallel to another given line, and `false` otherwise"""
        return are_lines_parallel(self, line)

    def is_coincident_to_line(self, line: "Line") -> bool:
        """Return `true` if the line is coincident to another given line, and `false` otherwise"""
        return are_lines_coincident(self, line)

    def get_intersection_with_line(self, line: "Line") -> list["Line"] | list["Point"]:
        """Compute and return the instersection of the line with another given line"""
        return get_intersection_of_lines(self, line)

    def get_distance_to_point(self, point: Point) -> float:
        """Compute and return the distance of a given point to the line"""
        return get_distance_between_line_and_point(self, point)

    def has_point(self, point: Point, max_distance: float = 0.01) -> bool:
        """Return `true` if a given point is on the line, `false` otherwise"""
        return line_has_point(self, point, max_distance)

    def transformed(self, transformation: "Transformation") -> "Line":
        """Return a copy of the line transformed with the given transformation"""
        return transform_line(self, transformation)

    def reversed(self) -> "Line":
        """Return a reversed copy of the line"""
        return reverse_line(self)

    def to_shapely(self) -> shapely.LineString:
        """Return a shapeply line string reproducing the line"""
        return shapely.LineString(
            [
                self.p1.to_tuple(),
                self.p2.to_tuple(),
            ]
        )


@dataclasses.dataclass(frozen=True)
class Segment(GeometryObject):
    """Class for segments"""

    p1: Point
    p2: Point

    def length(self) -> float:
        """Return the length of the segment"""
        return math.sqrt((self.p2.x - self.p1.x) ** 2 + (self.p2.y - self.p1.y) ** 2)

    def get_distance_to_point(self, point: Point) -> float:
        """Return the distance of a given point to the segment"""
        return get_distance_between_segment_and_point(self, point)

    def has_point(self, point: Point, max_distance: float = 0.01) -> bool:
        """Return `true` if the given point is on the segment, `false` otherwise"""
        return segment_has_point(self, point, max_distance)

    def get_angle_to_horizontal(self) -> float:
        """Compute and return the angle formed by the segment and the horizontal"""
        return get_angle_to_horizontal_of_line(self)

    def get_intersection_with_line(self, line: Line) -> list[Point] | list["Segment"]:
        """Compute and return the intersection of the segment with a given line"""
        return get_intersection_of_line_and_segment(line, self)

    def get_position_at_fraction(self, fraction: float) -> Point:
        """Compute and return the position on the segment at a given fraction (of the total length)"""
        return get_position_at_fraction_of_segment(self, fraction)

    def get_angle_at_fraction(self, fraction: float) -> float:
        """Compute and return the angle in radians formed by the segment and the horizontal at a given fraction (of the total length)"""
        return get_angle_at_fraction_of_segment(self, fraction)

    def get_position_and_angle_at_fraction(
        self, fraction: float
    ) -> tuple[Point, float]:
        """Compute and return the position on the segment at a given fraction and the angle in radians formed of the segment and the horizontal at that position"""
        return get_position_and_angle_at_fraction_of_segment(self, fraction)

    def shortened(
        self,
        length: float,
        start_or_end: typing.Literal["start", "end"] = "end",
    ) -> "Segment":
        """Compute and return a copy of the segment shortened by a given length"""
        return shorten_segment(self, length, start_or_end)

    def transformed(self, transformation: "Transformation") -> "Segment":
        """Compute and return a copy of the segment transformed with a given transformation"""
        return transform_segment(self, transformation)

    def reversed(self) -> "Segment":
        """Compute and return a reversed copy of the segment"""
        return reverse_segment(self)

    def to_shapely(self) -> shapely.LineString:
        """Return a shapely line string reproducing the segment"""
        return shapely.LineString(
            [
                self.p1.to_tuple(),
                self.p2.to_tuple(),
            ]
        )

    def bbox(self) -> "Bbox":
        """Compute and return the bounding box of the segment"""
        return Bbox.from_bounds(self.to_shapely().bounds)

    @classmethod
    def from_shapely(cls, line_string: shapely.LineString) -> typing_extensions.Self:
        """Compute and return the segment reproducing a shapely line string"""
        shapely_points = line_string.boundary.geoms
        return cls(
            Point.from_shapely(shapely_points[0]),
            Point.from_shapely(shapely_points[1]),
        )


@dataclasses.dataclass(frozen=True)
class BezierCurve(GeometryObject):
    """Class for bezier curves"""

    p1: Point
    p2: Point
    control_points: tuple[Point, ...] = dataclasses.field(default_factory=tuple)

    def _to_bezier(self):
        x = []
        y = []
        x.append(self.p1.x)
        y.append(self.p1.y)
        for point in self.control_points:
            x.append(point.x)
            y.append(point.y)
        x.append(self.p2.x)
        y.append(self.p2.y)
        nodes = [x, y]
        return bezier.curve.Curve.from_nodes(nodes)

    @classmethod
    def _from_bezier(cls, bezier_curve):
        points = [Point.from_tuple(t) for t in bezier_curve.nodes.T]
        return cls(points[0], points[-1], tuple(points[1:-1]))

    def length(self) -> float:
        """Compute and return the length of the bezier curve"""
        return self._to_bezier().length

    def evaluate(self, s: float) -> Point:
        """Compute and return the point at a given parameter value"""
        evaluation = self._to_bezier().evaluate(s)
        return Point.from_fortranarray(evaluation)

    def evaluate_multi(self, s: numpy.ndarray) -> list[Point]:
        """Compute and return the points at given parameter values"""
        evaluation = self._to_bezier().evaluate_multi(s)
        return [Point(e[0], e[1]) for e in evaluation.T]

    def get_intersection_with_line(self, line: Line) -> list[Point] | list[Segment]:
        """Compute and return the intersection of the bezier curve with a given line"""
        return get_intersection_of_line_and_bezier_curve(line, self)

    def get_position_at_fraction(self, fraction: float):
        """Compute and return the position on the bezier curve at a given fraction (of the total length)"""

        return get_position_at_fraction_of_bezier_curve(self, fraction)

    def get_angle_at_fraction(self, fraction: float):
        """Compute and return the angle in radians formed by the tangent of the bezier curve and the horizontal at a given fraction (of the total length)"""
        return get_angle_at_fraction_of_bezier_curve(self, fraction)

    def get_position_and_angle_at_fraction(self, fraction: float):
        """Compute and return the position on the bezier curve at a given fraction and the angle in radians formed of the tangent of the bezier curve and the horizontal at that position"""
        return get_position_and_angle_at_fraction_of_bezier_curve(self, fraction)

    def shortened(
        self, length: float, start_or_end: typing.Literal["start", "end"] = "end"
    ) -> "BezierCurve":
        """Compute and return a copy of the bezier curve shortened by a given length"""
        return shorten_bezier_curve(self, length, start_or_end)

    def transformed(self, transformation):
        """Compute and return a copy of the bezier curve transformed with a given transformation"""
        return transform_bezier_curve(self, transformation)

    def reversed(self):
        """Compute and return a reversed copy of the bezier curve"""
        return reverse_bezier_curve(self)

    def to_shapely(self, n_segs=50):
        """Compute and return a shapely line string reproducing the bezier curve"""

        points = self.evaluate_multi(
            numpy.arange(0, 1 + 1 / n_segs, 1 / n_segs, dtype="double")
        )
        return shapely.LineString([point.to_tuple() for point in points])

    def bbox(self):
        """Compute and return the bounding box of the bezier curve"""
        return Bbox.from_bounds(self.to_shapely().bounds)


@dataclasses.dataclass(frozen=True)
class EllipticalArc(GeometryObject):
    """Class for elliptical arcs"""

    p1: Point
    p2: Point
    rx: float
    ry: float
    x_axis_rotation: float
    arc_flag: int
    sweep_flag: int

    def get_intersection_with_line(self, line: Line) -> list[Point]:
        """Compute and return the intersection of the elliptical arc with a given line"""
        return get_intersection_of_line_and_elliptical_arc(line, self)

    def get_center_parameterization(
        self,
    ) -> tuple[float, float, float, float, float, float, float, float]:
        """Compute and return the center paramaterizaion of the elliptical arc (cx, cy, rx, ry, sigma, theta1, theta2, delta_theta)"""
        return get_center_parameterization_of_elliptical_arc(self)

    def get_center(self) -> Point:
        """Compute and return the center of the elliptical arc"""
        return get_center_of_elliptical_arc(self)

    def get_position_at_fraction(self, fraction: float) -> Point:
        """Compute and return the position on the elliptical arc at a given fraction (of the total length)"""
        return get_position_at_fraction_of_elliptical_arc(self, fraction)

    def get_angle_at_fraction(self, fraction: float) -> float:
        """Compute and return the angle in radians formed by the tangent of the elliptical arc and the horizontal at a given fraction (of the total length)"""
        return get_angle_at_fraction_of_elliptical_arc(self, fraction)

    def get_position_and_angle_at_fraction(
        self, fraction: float
    ) -> tuple[Point, float]:
        """Compute and return the position on the elliptical arc at a given fraction and the angle in radians formed of the tangent of the bezier curve at that position and the horizontal"""
        return get_position_and_angle_at_fraction_of_elliptical_arc(self, fraction)

    def to_shapely(self):
        """Compute and return a shapely linestring reproducing the elliptical arc"""

        def _split_line_string(
            line_string: shapely.LineString,
            point: Point,
        ):
            segment = Segment(
                Point.from_tuple(line_string.coords[0]),
                Point.from_tuple(line_string.coords[1]),
            )
            min_distance = segment.get_distance_to_point(point)
            min_i = 0
            for i, current_coord in enumerate(line_string.coords[2:]):
                previous_coord = line_string.coords[i + 1]
                segment = Segment(
                    Point.from_tuple(previous_coord),
                    Point.from_tuple(current_coord),
                )
                distance = segment.get_distance_to_point(point)
                if distance <= min_distance:
                    min_distance = distance
                    min_i = i
            left_coords = line_string.coords[0 : min_i + 1] + [point.to_shapely()]
            right_coords = [point.to_shapely()] + line_string.coords[min_i + 1 :]
            left_line_string = shapely.LineString(left_coords)
            right_line_string = shapely.LineString(right_coords)
            return [left_line_string, right_line_string]

        origin = shapely.Point(0, 0)
        circle = origin.buffer(1.0).boundary
        ellipse = shapely.affinity.scale(circle, self.rx, self.ry)
        ellipse = shapely.affinity.rotate(ellipse, self.x_axis_rotation)
        center = self.get_center()
        ellipse = shapely.affinity.translate(ellipse, center.x, center.y)
        line1 = Line(center, Point.from_tuple(ellipse.coords[0]))
        angle1 = get_angle_to_horizontal_of_line(line1)
        line2 = Line(center, Point.from_tuple(ellipse.coords[-2]))
        angle2 = get_angle_to_horizontal_of_line(line2)
        angle = angle1 - angle2
        if angle >= 0:
            sweep = 1
        else:
            sweep = 0
        if sweep != self.sweep_flag:
            ellipse = shapely.LineString(ellipse.coords[::-1])
        if ellipse.coords[0] == self.p1.to_tuple():
            ellipse = shapely.LineString(ellipse.coords[1:] + [ellipse.coords[0]])
        first_split = _split_line_string(ellipse, self.p1)
        multi_line = shapely.MultiLineString([first_split[1], first_split[0]])
        line_string = shapely.ops.linemerge(multi_line)
        second_split = _split_line_string(line_string, self.p2)
        shapely_arc = second_split[0]
        return shapely_arc

    def bbox(self) -> "Bbox":
        """Compute and return the bounding box of the elliptical arc"""
        return Bbox.from_bounds(self.to_shapely().bounds)

    def shortened(
        self,
        length: float,
        start_or_end: typing.Literal["start", "end"] = "end",
    ) -> "EllipticalArc":
        """Compute and return a copy of the elliptical arc shortened by a given length"""
        return shorten_elliptical_arc(self, length, start_or_end)

    def transformed(self, transformation: "Transformation") -> "EllipticalArc":
        """Compute and return a copy of the elliptical arc transformed by a given transformation"""
        return transform_elliptical_arc(self, transformation)

    def reversed(self) -> "EllipticalArc":
        """Compute and return a reversed copy of the elliptical arc"""
        return reverse_elliptical_arc(self)

    def to_bezier_curves(self) -> BezierCurve:
        """Compute and return a bezier curve reproducing the elliptical arc"""
        return transform_elliptical_arc_to_bezier_curves(self)

    def length(self) -> float:
        return self.to_shapely().length


@dataclasses.dataclass(frozen=True)
class Bbox(object):
    """Class for bounding boxes"""

    position: Point
    width: float
    height: float

    @property
    def x(self) -> float:
        """The x coordinate of the bounding box"""
        return self.position.x

    @property
    def y(self) -> float:
        """The y coordinate of the bounding box"""
        return self.position.y

    def size(self) -> tuple[float, float]:
        """The size of the bounding box"""
        return (self.width, self.height)

    def anchor_point(self, anchor_point: str) -> Point:
        """Return a given anchor point of the bounding box"""
        return getattr(self, anchor_point)()

    def north_west(self) -> Point:
        """Return the north west anchor point of the bounding box"""
        return Point(self.x - self.width / 2, self.y - self.height / 2)

    def north_north_west(self) -> Point:
        """Return the north north west anchor point of the bounding box"""
        return Point(self.x - self.width / 4, self.y - self.height / 2)

    def north(self) -> Point:
        """Return the north anchor point of the bounding box"""
        return Point(self.x, self.y - self.height / 2)

    def north_north_east(self) -> Point:
        """Return the north north east anchor point of the bounding box"""
        return Point(self.x + self.width / 4, self.y - self.height / 2)

    def north_east(self) -> Point:
        """Return the north east anchor point of the bounding box"""
        return Point(self.x + self.width / 2, self.y - self.height / 2)

    def east_north_east(self) -> Point:
        """Return the east north east anchor point of the bounding box"""
        return Point(self.x + self.width / 2, self.y - self.height / 4)

    def east(self) -> Point:
        """Return the east anchor point of the bounding box"""
        return Point(self.x + self.width / 2, self.y)

    def east_south_east(self) -> Point:
        """Return the east south east anchor point of the bounding box"""
        return Point(self.x + self.width / 2, self.y + self.width / 4)

    def south_east(self) -> Point:
        """Return the south east anchor point of the bounding box"""
        return Point(self.x + self.width / 2, self.y + self.height / 2)

    def south_south_east(self) -> Point:
        """Return the south south east anchor point of the bounding box"""
        return Point(self.x + self.width / 4, self.y + self.height / 2)

    def south(self) -> Point:
        """Return the south anchor point of the bounding box"""
        return Point(self.x, self.y + self.height / 2)

    def south_south_west(self) -> Point:
        """Return the south south west anchor point of the bounding box"""
        return Point(self.x - self.width / 4, self.y + self.height / 2)

    def south_west(self) -> Point:
        """Return the south west anchor point of the bounding box"""
        return Point(self.x - self.width / 2, self.y + self.height / 2)

    def west_south_west(self) -> Point:
        """Return the west south west anchor point of the bounding box"""
        return Point(self.x - self.width / 2, self.y + self.height / 4)

    def west(self) -> Point:
        """Return the west anchor point of the bounding box"""
        return Point(self.x - self.width / 2, self.y)

    def west_north_west(self) -> Point:
        """Return the west north west anchor point of the bounding box"""
        return Point(self.x - self.width / 2, self.y - self.height / 4)

    def center(self) -> Point:
        """Return the center anchor point of the bounding box"""
        return Point(self.x, self.y)

    def isnan(self) -> bool:
        """Return `true` if the position of the bounding box has a `nan` coordinate, `false` otherwise"""
        return self.position.isnan()

    @classmethod
    def from_bounds(cls, bounds: tuple[float, float, float, float]):
        """Create and return a bounding box from shaply bounds (min_x, min_y, max_x, max_y)"""
        return cls(
            Point((bounds[0] + bounds[2]) / 2, (bounds[1] + bounds[3]) / 2),
            bounds[2] - bounds[0],
            bounds[3] - bounds[1],
        )


@dataclasses.dataclass(frozen=True)
class Transformation(abc.ABC):
    """Base class for transformations"""

    @abc.abstractmethod
    def to_matrix(self) -> numpy.typing.NDArray:
        """Return a matrix representation of the transformation"""
        pass

    @abc.abstractmethod
    def inverted(self) -> "Transformation":
        """Compute and return the inverse transformation"""
        pass

    def __mul__(self, other):
        return MatrixTransformation(numpy.matmul(self.to_matrix(), other.to_matrix()))


@dataclasses.dataclass(frozen=True)
class MatrixTransformation(Transformation):
    """Class for matrix transformations"""

    m: numpy.typing.NDArray

    def to_matrix(self) -> numpy.typing.NDArray:
        """Return a matrix representation of the matrix transformation"""
        return self.m

    def inverted(self) -> Transformation:
        """Compute and return the inverse of the matrix transformation"""
        return invert_matrix_transformation(self)


@dataclasses.dataclass(frozen=True)
class Rotation(Transformation):
    """Class for rotations"""

    angle: float
    point: Point | None = None

    def to_matrix(self) -> numpy.typing.NDArray:
        """Compute and return a matrix representation of the rotation"""
        m = numpy.array(
            [
                [math.cos(self.angle), -math.sin(self.angle), 0],
                [math.sin(self.angle), math.cos(self.angle), 0],
                [0, 0, 1],
            ],
            dtype=float,
        )
        if self.point is not None:
            translation = Translation(self.point.x, self.point.y)
            m = numpy.matmul(
                numpy.matmul(translation.to_matrix(), m),
                translation.inverted().to_matrix(),
            )
        return m

    def inverted(self) -> Transformation:
        """Compute and return the inverse of the rotation"""
        return invert_rotation(self)


@dataclasses.dataclass(frozen=True)
class Translation(Transformation):
    """Class for translations"""

    tx: float
    ty: float

    def to_matrix(self) -> numpy.typing.NDArray:
        """Return a matrix representation of the translation"""
        m = numpy.array([[1, 0, self.tx], [0, 1, self.ty], [0, 0, 1]], dtype=float)
        return m

    def inverted(self) -> Transformation:
        """Compute and return the inverse of the translation"""
        return invert_translation(self)


@dataclasses.dataclass(frozen=True)
class Scaling(Transformation):
    """Class for scalings"""

    sx: float
    sy: float

    def to_matrix(self) -> numpy.typing.NDArray:
        """Return a matrix representation of the scaling"""
        m = numpy.array(
            [
                [self.sx, 0, 0],
                [0, self.sy, 0],
                [0, 0, 1],
            ],
            dtype=float,
        )
        return m

    def inverted(self) -> Transformation:
        """Compute and return the inverse of the scaling"""
        return invert_scaling(self)


def transform_point(point: Point, transformation: Transformation) -> Point:
    m = numpy.matmul(transformation.to_matrix(), point.to_matrix())
    return Point(m[0][0], m[1][0])


def transform_line(line: Line, transformation: Transformation) -> Line:
    return Line(
        transform_point(line.p1, transformation),
        transform_point(line.p2, transformation),
    )


def transform_segment(segment: Segment, transformation: Transformation) -> Segment:
    return Segment(
        transform_point(segment.p1, transformation),
        transform_point(segment.p2, transformation),
    )


def transform_bezier_curve(
    bezier_curve: BezierCurve, transformation: Transformation
) -> BezierCurve:
    return BezierCurve(
        transform_point(bezier_curve.p1, transformation),
        transform_point(bezier_curve.p2, transformation),
        tuple(
            [
                transform_point(point, transformation)
                for point in bezier_curve.control_points
            ]
        ),
    )


def transform_elliptical_arc(
    elliptical_arc: EllipticalArc, transformation: Transformation
) -> EllipticalArc:
    east = Point(
        math.cos(elliptical_arc.x_axis_rotation) * elliptical_arc.rx,
        math.sin(elliptical_arc.x_axis_rotation) * elliptical_arc.rx,
    )
    north = Point(
        math.cos(elliptical_arc.x_axis_rotation) * elliptical_arc.ry,
        math.sin(elliptical_arc.x_axis_rotation) * elliptical_arc.ry,
    )
    new_center = transform_point(Point(0, 0), transformation)
    new_east = transform_point(east, transformation)
    new_north = transform_point(north, transformation)
    new_rx = Segment(new_center, new_east).length()
    new_ry = Segment(new_center, new_north).length()
    new_start_point = transform_point(elliptical_arc.p1, transformation)
    new_end_point = transform_point(elliptical_arc.p2, transformation)
    new_x_axis_rotation = math.degrees(
        get_angle_to_horizontal_of_line(Line(new_center, new_east))
    )
    return EllipticalArc(
        p1=new_start_point,
        p2=new_end_point,
        rx=new_rx,
        ry=new_ry,
        x_axis_rotation=new_x_axis_rotation,
        arc_flag=elliptical_arc.arc_flag,
        sweep_flag=elliptical_arc.sweep_flag,
    )


def reverse_point(point: Point) -> Point:
    return Point(point.x, point.y)


def reverse_line(line: Line) -> Line:
    return Line(line.p2, line.p1)


def reverse_segment(segment):
    return Segment(segment.p2, segment.p1)


def reverse_bezier_curve(bezier_curve: BezierCurve) -> BezierCurve:
    return BezierCurve(
        bezier_curve.p2, bezier_curve.p1, bezier_curve.control_points[::-1]
    )


def reverse_elliptical_arc(elliptical_arc: EllipticalArc) -> EllipticalArc:
    return EllipticalArc(
        elliptical_arc.p2,
        elliptical_arc.p1,
        elliptical_arc.rx,
        elliptical_arc.ry,
        elliptical_arc.x_axis_rotation,
        elliptical_arc.arc_flag,
        abs(elliptical_arc.sweep_flag - 1),
    )


def shorten_segment(
    segment: Segment,
    length: float,
    start_or_end: typing.Literal["start", "end"] = "end",
) -> Segment:
    if length == 0 or segment.length() == 0:
        shortened_segment = copy.deepcopy(segment)
    else:
        if start_or_end == "start":
            shortened_segment = segment.reversed().shortened(length).reversed()
        else:
            fraction = 1 - length / segment.length()
            point = segment.get_position_at_fraction(fraction)
            shortened_segment = Segment(segment.p1, point)
    return shortened_segment


def shorten_bezier_curve(
    bezier_curve: BezierCurve,
    length: float,
    start_or_end: typing.Literal["start", "end"] = "end",
) -> BezierCurve:
    if length == 0 or bezier_curve.length() == 0:
        shortened_bezier_curve = copy.deepcopy(bezier_curve)
    else:
        if start_or_end == "start":
            shortened_bezier_curve = (
                bezier_curve.reversed().shortened(length).reversed()
            )
        else:
            lib_bezier_curve = bezier_curve._to_bezier()
            total_length = lib_bezier_curve.length
            if length > total_length:
                length = total_length
            fraction = 1 - length / total_length
            point = bezier_curve.get_position_at_fraction(fraction)
            horizontal_line = BezierCurve(point - (5, 0), point + (5, 0))._to_bezier()
            s = lib_bezier_curve.intersect(horizontal_line)[0][0]
            lib_shortened_bezier_curve = lib_bezier_curve.specialize(0, s)
            shortened_bezier_curve = BezierCurve._from_bezier(
                lib_shortened_bezier_curve
            )
    return shortened_bezier_curve


def shorten_elliptical_arc(
    elliptical_arc: EllipticalArc,
    length: float,
    start_or_end: typing.Literal["start", "end"] = "end",
):
    if length == 0 or elliptical_arc.length() == 0:
        shortened_elliptical_arc = copy.deepcopy(elliptical_arc)
    else:
        if start_or_end == "start":
            shortened_elliptical_arc = (
                elliptical_arc.reversed().shortened(length).reversed()
            )
        else:
            fraction = 1 - length / elliptical_arc.length()
            point = elliptical_arc.get_position_at_fraction(fraction)
            shortened_elliptical_arc = dataclasses.replace(elliptical_arc, p2=point)
    return shortened_elliptical_arc


def invert_matrix_transformation(
    matrix_transformation: MatrixTransformation,
) -> MatrixTransformation:
    return MatrixTransformation(numpy.linalg.inv(matrix_transformation.m))


def invert_rotation(rotation: Rotation) -> Rotation:
    return Rotation(-rotation.angle, rotation.point)


def invert_translation(translation: Translation) -> Translation:
    return Translation(-translation.tx, -translation.ty)


def invert_scaling(scaling: Scaling) -> Scaling:
    return Scaling(-scaling.sx, -scaling.sy)


def get_intersection_of_line_and_point(line: Line, point: Point) -> list[Point]:
    if line.has_point(point):
        return [point]
    return []


def get_intersection_of_lines(line1: Line, line2: Line) -> list[Line] | list[Point]:
    slope1 = line1.slope()
    intercept1 = line1.intercept()
    slope2 = line2.slope()
    intercept2 = line2.intercept()
    if line1.is_coincident_to_line(line2):
        intersection = [copy.deepcopy(line1)]
    elif line1.is_parallel_to_line(line2):
        intersection = []
    elif math.isnan(slope1):
        intersection = [Point(line1.p1.x, slope2 * line1.p1.x + intercept2)]
    elif math.isnan(slope2):
        intersection = [Point(line2.p1.x, slope1 * line2.p1.x + intercept1)]
    else:
        d = (line1.p1.x - line1.p2.x) * (line2.p1.y - line2.p2.y) - (
            line1.p1.y - line1.p2.y
        ) * (line2.p1.x - line2.p2.x)
        px = (
            (line1.p1.x * line1.p2.y - line1.p1.y * line1.p2.x)
            * (line2.p1.x - line2.p2.x)
            - (line1.p1.x - line1.p2.x)
            * (line2.p1.x * line2.p2.y - line2.p1.y * line2.p2.x)
        ) / d
        py = (
            (line1.p1.x * line1.p2.y - line1.p1.y * line1.p2.x)
            * (line2.p1.y - line2.p2.y)
            - (line1.p1.y - line1.p2.y)
            * (line2.p1.x * line2.p2.y - line2.p1.y * line2.p2.x)
        ) / d
        intersection = [Point(px, py)]
    return intersection


def get_intersection_of_line_and_segment(
    line: Line, segment: Segment
) -> list[Point] | list[Segment] | list[Line]:
    line2 = Line(segment.p1, segment.p2)
    intersection = line.get_intersection_with_line(line2)
    if len(intersection) > 0 and isinstance(intersection[0], Point):
        sorted_xs = sorted([segment.p1.x, segment.p2.x])
        sorted_ys = sorted([segment.p1.y, segment.p2.y])
        if not (
            intersection[0].x >= sorted_xs[0]
            and intersection[0].x <= sorted_xs[-1]
            and intersection[0].y >= sorted_ys[0]
            and intersection[0].y <= sorted_ys[-1]
        ):
            intersection = []
    elif len(intersection) > 0:
        intersection = [segment]
    return intersection


def get_intersection_of_line_and_bezier_curve(
    line: Line, bezier_curve: BezierCurve
) -> list[Point] | list[Segment]:
    shapely_object = bezier_curve.to_shapely()
    return get_intersection_of_line_and_shapely_object(line, shapely_object)


def get_intersection_of_line_and_elliptical_arc(
    line: Line, elliptical_arc: EllipticalArc
) -> list[Point] | list[Segment]:
    shapely_object = elliptical_arc.to_shapely()
    return get_intersection_of_line_and_shapely_object(line, shapely_object)


def get_intersection_of_line_and_shapely_object(
    line: Line, shapely_object: shapely.Geometry
) -> list[Point] | list[Segment]:
    intersection = []
    for shapely_geom in shapely_object.geoms:
        bbox = Bbox.from_bounds(shapely_object.bounds)
        slope = line.slope()
        north_west = bbox.north_west()
        south_east = bbox.south_east()
        offset = 100.0
        if not math.isnan(slope):
            intercept = line.intercept()
            left_x = north_west.x - offset
            left_y = slope * left_x + intercept
            right_x = south_east.x + offset
            right_y = slope * right_x + intercept
        else:
            left_x = line.p1.x
            left_y = north_west.y - offset
            right_x = line.p1.x
            right_y = south_east.y + offset
        left_point = Point(left_x, left_y)
        right_point = Point(right_x, right_y)
        line_string = shapely.LineString(
            [left_point.to_shapely(), right_point.to_shapely()]
        )
        shapely_intersection = line_string.intersection(shapely_geom)
        if not hasattr(shapely_intersection, "geoms"):
            shapely_intersection = [shapely_intersection]
        else:
            shapely_intersection = shapely_intersection.geoms
        for shapely_obj in shapely_intersection:
            if not shapely.is_empty(shapely_obj):
                if isinstance(shapely_obj, shapely.Point):
                    intersection_obj = Point.from_shapely(shapely_obj)
                elif isinstance(shapely_obj, shapely.LineString):
                    intersection_obj = Segment.from_shapely(shapely_obj)
                intersection.append(intersection_obj)
    return intersection


def get_shapely_object_bbox(shapely_object: shapely.Geometry) -> Bbox:
    return Bbox.from_bounds(shapely_object.bounds)


def get_shapely_object_border(
    shapely_object: shapely.Geometry, point: Point, center: Point | None = None
) -> Point | None:
    if center is None:
        bbox = get_shapely_object_bbox(shapely_object)
        center = bbox.center()
    if center.isnan():
        return Point(float("nan"), float("nan"))
    line = Line(center, point)
    intersection = get_intersection_of_line_and_shapely_object(line, shapely_object)
    candidate_points = []
    for intersection_obj in intersection:
        if isinstance(intersection_obj, Segment):
            candidate_points.append(intersection_obj.p1)
            candidate_points.append(intersection_obj.p2)
        elif isinstance(intersection_obj, Point):
            candidate_points.append(intersection_obj)
    intersection_point = None
    max_d = -1
    ok_direction_exists = False
    d1 = get_distance_between_points(point, center)
    for candidate_point in candidate_points:
        d2 = get_distance_between_points(candidate_point, point)
        d3 = get_distance_between_points(candidate_point, center)
        candidate_ok_direction = not d2 > d1 or d2 < d3
        if candidate_ok_direction or not ok_direction_exists:
            if candidate_ok_direction and not ok_direction_exists:
                ok_direction_exists = True
                max_d = -1
            if d3 > max_d:
                max_d = d3
                intersection_point = candidate_point
    return intersection_point


def get_shapely_object_angle(
    shapely_object: shapely.Geometry,
    angle: float,
    unit: typing.Literal["degrees", "radians"] = "degrees",
    center: Point | None = None,
) -> Point | None:
    if unit == "degrees":
        angle = math.radians(angle)
    angle = -angle
    d = 100
    if center is None:
        bbox = get_shapely_object_bbox(shapely_object)
        center = bbox.center()
        if center.isnan():
            return Point(float("nan"), float("nan"))
    point = center + (d * math.cos(angle), d * math.sin(angle))
    return get_shapely_object_border(shapely_object, point, center)


def get_shapely_object_anchor_point(
    shapely_object: shapely.Geometry,
    anchor_point: str,
    center: Point | None = None,
) -> Point:
    bbox = get_shapely_object_bbox(shapely_object)
    if center is None:
        center = bbox.center()
    if center.isnan():
        return Point(float("nan"), float("nan"))
    point = bbox.anchor_point(anchor_point)
    return get_shapely_object_border(shapely_object, point, center)


def get_angle_to_horizontal_of_point(point: Point) -> float:
    line = Line(Point(0, 0), point)
    return line.get_angle_to_horizontal()


# angle in radians
def get_angle_to_horizontal_of_line(line: Line | Segment) -> float:
    x1 = line.p1.x
    y1 = line.p1.y
    x2 = line.p2.x
    y2 = line.p2.y
    angle = math.atan2(y2 - y1, x2 - x1)
    return get_normalized_angle(angle)


def are_lines_parallel(line1: Line, line2: Line) -> bool:
    slope1 = line1.slope()
    slope2 = line2.slope()
    if math.isnan(slope1) and math.isnan(slope2):
        return True
    return slope1 == slope2


def are_lines_coincident(line1: Line, line2: Line) -> bool:
    slope1 = line1.slope()
    intercept1 = line1.intercept()
    slope2 = line2.slope()
    intercept2 = line2.intercept()
    return (
        math.isnan(slope1)
        and math.isnan(slope2)
        and line1.p1.x == line2.p1.x
        or slope1 == slope2
        and intercept1 == intercept2
    )


# angle in radians
def is_angle_in_sector(
    angle: float, center: Point, point1: Point, point2: Point
) -> bool:
    angle1 = get_angle_to_horizontal_of_line(Line(center, point1))
    angle2 = get_angle_to_horizontal_of_line(Line(center, point2))
    return is_angle_between(angle, angle1, angle2)


# angles in radians
def is_angle_between(angle: float, start_angle: float, end_angle: float) -> bool:
    angle = get_normalized_angle(angle)
    start_angle = get_normalized_angle(start_angle)
    end_angle = get_normalized_angle(end_angle)
    if start_angle < end_angle:
        if angle >= start_angle and angle <= end_angle:
            return True
    else:
        if start_angle <= angle <= 2 * math.pi or angle >= 0 and angle <= end_angle:
            return True
    return False


# angle is in radians; return angle between 0 and 2 * pi
def get_normalized_angle(angle: float) -> float:
    return angle - (angle // (2 * math.pi) * (2 * math.pi))


def _get_position_at_fraction(segment_or_curve, fraction):
    line_string = segment_or_curve.to_shapely()
    shapely_point = line_string.interpolate(fraction, normalized=True)
    return Point.from_shapely(shapely_point)


def get_position_at_fraction_of_segment(
    segment: Segment,
    fraction: float,
) -> Point:  # fraction in [0, 1]
    return _get_position_at_fraction(segment, fraction)


def get_position_at_fraction_of_bezier_curve(
    bezier_curve: BezierCurve,
    fraction: float,
) -> Point:  # fraction in [0, 1]
    return _get_position_at_fraction(bezier_curve, fraction)


def get_position_at_fraction_of_elliptical_arc(
    elliptical_arc: EllipticalArc,
    fraction: float,
) -> Point:  # fraction in [0, 1]
    return _get_position_at_fraction(elliptical_arc, fraction)


def get_center_parameterization_of_elliptical_arc(
    elliptical_arc: EllipticalArc,
) -> tuple[float, float, float, float, float, float, float, float]:
    x1, y1 = elliptical_arc.p1.x, elliptical_arc.p1.y
    sigma = elliptical_arc.x_axis_rotation
    x2, y2 = elliptical_arc.p2.x, elliptical_arc.p2.y
    rx = elliptical_arc.rx
    ry = elliptical_arc.ry
    fa = elliptical_arc.arc_flag
    fs = elliptical_arc.sweep_flag
    x1p = math.cos(sigma) * ((x1 - x2) / 2) + math.sin(sigma) * ((y1 - y2) / 2)
    y1p = -math.sin(sigma) * ((x1 - x2) / 2) + math.cos(sigma) * ((y1 - y2) / 2)
    l = x1p**2 / rx**2 + y1p**2 / ry**2
    if l > 1:
        rx = math.sqrt(l) * rx
        ry = math.sqrt(l) * ry
    r = rx**2 * ry**2 - rx**2 * y1p**2 - ry**2 * x1p**2
    if r < 0:  # due to imprecision? to fix later
        r = 0
    a = math.sqrt(r / (rx**2 * y1p**2 + ry**2 * x1p**2))
    if fa == fs:
        a = -a
    cxp = a * rx * y1p / ry
    cyp = -a * ry * x1p / rx
    cx = math.cos(sigma) * cxp - math.sin(sigma) * cyp + (x1 + x2) / 2
    cy = math.sin(sigma) * cxp + math.cos(sigma) * cyp + (y1 + y2) / 2
    theta1 = get_angle_between_segments(
        Segment(Point(0, 0), Point(1, 0)),
        Segment(Point(0, 0), Point((x1p - cxp) / rx, (y1p - cyp) / ry)),
    )
    delta_theta = get_angle_between_segments(
        Segment(Point(0, 0), Point((x1p - cxp) / rx, (y1p - cyp) / ry)),
        Segment(Point(0, 0), Point(-(x1p + cxp) / rx, -(y1p + cyp) / ry)),
    )
    if fs == 0 and delta_theta > 0:
        delta_theta -= 2 * math.pi
    elif fs == 1 and delta_theta < 0:
        delta_theta += 2 * math.pi
    theta2 = theta1 + delta_theta
    return cx, cy, rx, ry, sigma, theta1, theta2, delta_theta


def get_center_of_elliptical_arc(elliptical_arc: EllipticalArc) -> Point:
    cx, cy, *_ = get_center_parameterization_of_elliptical_arc(elliptical_arc)
    return Point(cx, cy)


def transform_elliptical_arc_to_bezier_curves(
    elliptical_arc: EllipticalArc,
) -> list[BezierCurve]:
    def _make_angles(angles):
        new_angles = [angles[0]]
        for theta1, theta2 in zip(angles, angles[1:]):
            delta_theta = theta2 - theta1
            if delta_theta > math.pi / 2:
                new_angles.append(theta1 + delta_theta / 2)
            new_angles.append(theta2)
        if len(new_angles) != len(angles):
            new_angles = _make_angles(new_angles)
        return new_angles

    bezier_curves = []
    (
        cx,
        cy,
        rx,
        ry,
        sigma,
        theta1,
        theta2,
        delta_theta,
    ) = get_center_parameterization_of_elliptical_arc(elliptical_arc)
    translation = Translation(cx, cy)
    scaling = Scaling(rx, ry)
    rotation = Rotation(sigma)
    transformation = translation * rotation * scaling
    angles = _make_angles([theta1, theta2])
    for theta1, theta2 in zip(angles, angles[1:]):
        x1 = math.cos(theta1)
        y1 = math.sin(theta1)
        x2 = math.cos(theta2)
        y2 = math.sin(theta2)
        ax = x1
        ay = y1
        bx = x2
        by = y2
        q1 = ax * ax + ay * ay
        q2 = q1 + ax * bx + ay * by
        k2 = (4 / 3) * (math.sqrt(2 * q1 * q2) - q2) / (ax * by - ay * bx)
        cp1x = ax - k2 * ay
        cp1y = ay + k2 * ax
        cp2x = bx + k2 * by
        cp2y = by - k2 * bx
        p1 = Point(x1, y1)
        p2 = Point(x2, y2)
        control_point1 = Point(cp1x, cp1y)
        control_point2 = Point(cp2x, cp2y)
        bezier_curve = BezierCurve(
            p1=p1, p2=p2, control_points=tuple([control_point1, control_point2])
        ).transformed(transformation)
        bezier_curves.append(bezier_curve)
    return bezier_curves


def _get_angle_at_fraction(
    segment_or_curve: Segment | BezierCurve | EllipticalArc,
    fraction: float,
) -> float:  # fraction in [0, 1]
    line_string = segment_or_curve.to_shapely()
    total_length = line_string.length
    current_length = 0
    previous_coord = line_string.coords[0]
    for current_coord in line_string.coords[1:]:
        segment = Segment(
            Point.from_tuple(previous_coord),
            Point.from_tuple(current_coord),
        )
        current_length += segment.length()
        if current_length / total_length >= fraction:
            break
    return segment.get_angle_to_horizontal()


def get_angle_at_fraction_of_segment(
    segment: Segment,
    fraction: float,
) -> float:  # fraction in [0, 1]
    return _get_angle_at_fraction(segment, fraction)


def get_angle_at_fraction_of_bezier_curve(
    bezier_curve: BezierCurve,
    fraction: float,
) -> float:  # fraction in [0, 1]
    return _get_angle_at_fraction(bezier_curve, fraction)


def get_angle_at_fraction_of_elliptical_arc(
    elliptical_arc: EllipticalArc,
    fraction: float,
) -> float:  # fraction in [0, 1]
    return _get_angle_at_fraction(elliptical_arc, fraction)


def _get_position_and_angle_at_fraction(
    segment_or_curve: Segment | BezierCurve | EllipticalArc,
    fraction: float,
) -> tuple[Point, float]:  # fraction in [0, 1]
    position = _get_position_at_fraction(segment_or_curve, fraction)
    angle = _get_angle_at_fraction(segment_or_curve, fraction)
    return position, angle


def get_position_and_angle_at_fraction_of_segment(
    segment: Segment,
    fraction: float,
) -> tuple[Point, float]:  # fraction in [0, 1]
    return _get_position_and_angle_at_fraction(segment, fraction)


def get_position_and_angle_at_fraction_of_bezier_curve(
    bezier_curve: BezierCurve,
    fraction: float,
) -> tuple[Point, float]:  # fraction in [0, 1]
    return _get_position_and_angle_at_fraction(bezier_curve, fraction)


def get_position_and_angle_at_fraction_of_elliptical_arc(
    elliptical_arc: EllipticalArc,
    fraction: float,
) -> tuple[Point, float]:  # fraction in [0, 1]
    return _get_position_and_angle_at_fraction(elliptical_arc, fraction)


# angle is in radians between -pi and pi
def get_angle_between_segments(segment1: Segment, segment2: Segment) -> float:
    p1 = segment1.p2 - segment1.p1
    p2 = segment2.p2 - segment2.p1
    scalar_prod = p1.x * p2.x + p1.y * p2.y
    angle = math.acos(
        round(scalar_prod / (segment1.length() * segment2.length()), ROUNDING)
    )
    sign = p1.x * p2.y - p1.y * p2.x
    if sign < 0:
        angle = -angle
    return angle


def get_distance_between_points(p1: Point, p2: Point) -> float:
    return Segment(p1, p2).length()


def get_distance_between_line_and_point(line: Line, point: Point) -> float:
    distance = abs(
        (line.p2.x - line.p1.x) * (line.p1.y - point.y)
        - (line.p1.x - point.x) * (line.p2.y - line.p1.y)
    ) / math.sqrt((line.p2.x - line.p1.x) ** 2 + (line.p2.y - line.p1.y) ** 2)
    return distance


def get_distance_between_segment_and_point(segment: Segment, point: Point) -> float:
    a = point.x - segment.p1.x
    b = point.y - segment.p1.y
    c = segment.p2.x - segment.p1.x
    d = segment.p2.y - segment.p1.y
    dot = a * c + b * d
    len_sq = c**2 + d**2
    if len_sq != 0:
        param = dot / len_sq
    else:
        param = -1
    if param < 0:
        xx = segment.p1.x
        yy = segment.p1.y
    elif param > 1:
        xx = segment.p2.x
        yy = segment.p2.y
    else:
        xx = segment.p1.x + param * c
        yy = segment.p1.y + param * d
    dx = point.x - xx
    dy = point.y - yy
    distance = math.sqrt(dx**2 + dy**2)
    return distance


def line_has_point(line: Line, point: Point, max_distance: float = 0.01) -> bool:
    d = line.get_distance_to_point(point)
    if d <= max_distance:
        return True
    return False


def segment_has_point(
    segment: Segment, point: Point, max_distance: float = 0.01
) -> bool:
    d = segment.get_distance_to_point(point)
    if d <= max_distance:
        return True
    return False


# Given a frame F defined in another reference frame by its origin,
# unit x axis vector, and unit y axis vector, returns the transformation
# that must be applied to a point defined in F to obtain its coordinates
# in the reference frame.
def get_transformation_for_frame(
    origin: Point, unit_x: Point, unit_y: Point
) -> MatrixTransformation:
    m = numpy.array(
        [
            [
                unit_x.x - origin.x,
                unit_y.x - origin.x,
                origin.x,
            ],
            [
                unit_x.y - origin.y,
                unit_y.y - origin.y,
                origin.y,
            ],
            [0, 0, 1],
        ],
        dtype=float,
    )
    return MatrixTransformation(m)


def _point_builder_add(self, xy):
    if momapy.builder.isinstance_or_builder(xy, Point):
        xy = (xy.x, xy.y)
    return PointBuilder(self.x + xy[0], self.y + xy[1])


def _point_builder_sub(self, xy):
    if momapy.builder.isinstance_or_builder(xy, Point):
        xy = (xy.x, xy.y)
    return PointBuilder(self.x - xy[0], self.y - xy[1])


def _point_builder_mul(self, scalar):
    return PointBuilder(self.x * scalar, self.y * scalar)


def _point_builder_div(self, scalar):
    return PointBuilder(self.x / scalar, self.y / scalar)


def _point_builder_iter(self):
    yield self.x
    yield self.y


def _point_builder_post_init(self):
    object.__setattr__(self, "x", round(self.x, ROUNDING))
    object.__setattr__(self, "y", round(self.y, ROUNDING))


PointBuilder = momapy.builder.get_or_make_builder_cls(
    Point,
    builder_namespace={
        "__add__": _point_builder_add,
        "__sub__": _point_builder_sub,
        "__mul__": _point_builder_mul,
        "__div__": _point_builder_div,
        "__iter__": _point_builder_iter,
        "__post_init__": _point_builder_post_init,
    },
)

momapy.builder.register_builder_cls(PointBuilder)

BboxBuilder = momapy.builder.get_or_make_builder_cls(Bbox)

momapy.builder.register_builder_cls(BboxBuilder)
